# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import json
import copy
import logging
from flask_babel import _
from superdesk.resource import Resource
from superdesk.services import BaseService
from superdesk.utils import ListCursor
from superdesk import get_resource_service, config, app
from superdesk.io.subjectcodes import get_subjectcodeitems
from eve.utils import ParsedRequest


logger = logging.getLogger(__name__)


class FilterConditionParametersResource(Resource):
    url = "filter_conditions/parameters"
    resource_methods = ['GET']
    item_methods = []
    schema = {
        'field': {'type': 'string'},
        'label': {'type': 'string'},
        'operators': {'type': 'list'},
        'values': {'type': 'list'},
        'value_field': {'type': 'string'},
    }


class FilterConditionParametersService(BaseService):
    def get(self, req, lookup):
        values = self._get_field_values()
        fields = [{'field': 'anpa_category',
                   'label': _('ANPA Category'),
                   'operators': ['in', 'nin'],
                   'values': values.get('anpa_category', []),
                   'value_field': 'qcode'
                   },
                  {'field': 'urgency',
                   'label': _('Urgency'),
                   'operators': ['in', 'nin', 'eq', 'ne', 'lt', 'lte', 'gt', 'gte'],
                   'values': values.get('urgency', []),
                   'value_field': 'qcode'
                   },
                  {'field': 'genre',
                   'label': _('Genre'),
                   'operators': ['in', 'nin'],
                   'values': values.get('genre', []),
                   'value_field': 'qcode'
                   },
                  {'field': 'subject',
                   'label': _('Subject'),
                   'operators': ['in', 'nin'],
                   'values': values.get('subject', []),
                   'value_field': 'qcode'
                   },
                  {'field': 'priority',
                   'label': _('Priority'),
                   'operators': ['in', 'nin', 'eq', 'ne', 'lt', 'lte', 'gt', 'gte'],
                   'values': values.get('priority', []),
                   'value_field': 'qcode'
                   },
                  {'field': 'keywords',
                   'label': _('Keywords'),
                   'operators': ['in', 'nin']
                   },
                  {'field': 'slugline',
                   'label': _('Slugline'),
                   'operators': ['in', 'nin', 'eq', 'ne', 'like', 'notlike', 'startswith', 'endswith']
                   },
                  {'field': 'type',
                   'label': _('Type'),
                   'operators': ['in', 'nin', 'eq', 'ne'],
                   'values': values.get('type', []),
                   'value_field': 'qcode'
                   },
                  {'field': 'source',
                   'label': _('Source'),
                   'operators': ['in', 'nin', 'eq', 'ne', 'like', 'notlike', 'startswith', 'endswith']
                   },
                  {'field': 'headline',
                   'label': _('Headline'),
                   'operators': ['in', 'nin', 'eq', 'ne', 'like', 'notlike', 'startswith', 'endswith']
                   },
                  {'field': 'ednote',
                   'label': _('Ednote'),
                   'operators': ['in', 'nin', 'eq', 'ne', 'like', 'notlike', 'startswith', 'endswith']
                   },
                  {'field': 'body_html',
                   'label': _('Body HTML'),
                   'operators': ['in', 'nin', 'like', 'notlike', 'startswith', 'endswith']
                   },
                  {'field': 'desk',
                   'label': _('Desk'),
                   'operators': ['in', 'nin', 'eq', 'ne'],
                   'values': values['desk'],
                   'value_field': '_id'
                   },
                  {'field': 'stage',
                   'label': _('Stage'),
                   'operators': ['in', 'nin', 'eq', 'ne'],
                   'values': values['stage'],
                   'value_field': '_id'
                   },
                  {'field': 'sms',
                   'label': _('SMS'),
                   'operators': ['in', 'nin', 'eq', 'ne'],
                   'values': values['sms'],
                   'value_field': 'name'
                   },
                  {'field': 'place',
                   'label': _('Place'),
                   'operators': ['match', 'in', 'nin'],
                   'values': values['place'],
                   'value_field': 'qcode'
                   },
                  {'field': 'ingest_provider',
                   'label': _('Ingest provider'),
                   'operators': ['eq', 'ne'],
                   'values': values['ingest_provider'],
                   'value_field': '_id'
                   },
                  {'field': 'embargo',
                   'label': _('Embargo'),
                   'operators': ['eq', 'ne'],
                   'values': values['embargo'],
                   'value_field': 'name'
                   },
                  {'field': 'featuremedia',
                   'label': _('Feature Media'),
                   'operators': ['exists'],
                   'values': values['featuremedia'],
                   'value_field': 'name'
                   },
                  {'field': 'anpa_take_key',
                   'operators': ['in', 'nin', 'eq', 'ne', 'like', 'notlike', 'startswith', 'endswith']
                   },
                  ]

        if 'planning' in app.config.get('INSTALLED_APPS', []):
            fields.append({'field': 'agendas',
                           'label': _('Agendas'),
                           'operators': ['in', 'nin'],
                           'values': list(get_resource_service('agenda').find({'is_enabled': True})),
                           'value_field': '_id',
                           })

        fields.extend(self._get_vocabulary_fields(values))
        return ListCursor(fields)

    def _get_vocabulary_fields(self, values):
        excluded_vocabularies = copy.copy(app.config.get('EXCLUDED_VOCABULARY_FIELDS', []))
        excluded_vocabularies.extend(values)
        lookup = {'_id': {'$nin': excluded_vocabularies}, 'type': 'manageable'}
        for vocabulary in get_resource_service('vocabularies').get(req=None, lookup=lookup):
            field = {'field': vocabulary[config.ID_FIELD], 'label': vocabulary['display_name']}

            if vocabulary.get('field_type') and vocabulary.get('field_type', '') != 'text':
                continue

            if vocabulary.get('field_type', '') == 'text':
                field['operators'] = ['in', 'nin', 'eq', 'ne', 'like', 'notlike', 'startswith', 'endswith']
            else:
                field['values'] = vocabulary['items']
                field['operators'] = ['in', 'nin']
                field['value_field'] = 'qcode'
            yield field

    def _get_field_values(self):
        values = {}
        vocabularies_resource = get_resource_service('vocabularies')
        categories_cv = vocabularies_resource.find_one(req=None, _id='categories')
        values['anpa_category'] = categories_cv.get('items') if categories_cv else []
        req = ParsedRequest()
        req.where = json.dumps({'$or': [{"schema_field": "genre"}, {"_id": "genre"}]})
        genre = vocabularies_resource.get(req=req, lookup=None)
        if genre.count():
            values['genre'] = genre[0]['items']
        for voc_id in ('urgency', 'priority', 'type'):
            try:
                values[voc_id] = vocabularies_resource.find_one(req=None, _id=voc_id)['items']
            except TypeError:
                values[voc_id] = []
        subject = vocabularies_resource.find_one(req=None, schema_field='subject')
        if subject:
            values['subject'] = subject['items']
        else:
            values['subject'] = get_subjectcodeitems()
        values['desk'] = list(get_resource_service('desks').get(None, {}))
        values['stage'] = self._get_stage_field_values(values['desk'])
        values['sms'] = [{'qcode': 0, 'name': 'False'}, {'qcode': 1, 'name': 'True'}]
        values['embargo'] = [{'qcode': 0, 'name': 'False'}, {'qcode': 1, 'name': 'True'}]
        req = ParsedRequest()
        req.where = json.dumps({'$or': [{"schema_field": "place"}, {"_id": "place"}, {"_id": "locators"}]})
        place = vocabularies_resource.get(req=req, lookup=None)
        if place.count():
            values['place'] = place[0]['items']
        else:
            values['place'] = []
        values['ingest_provider'] = list(get_resource_service('ingest_providers').get(None, {}))
        values['featuremedia'] = [{'qcode': 1, 'name': 'True'}, {'qcode': 0, 'name': 'False'}]
        return values

    def _get_stage_field_values(self, desks):
        stages = list(get_resource_service('stages').get(None, {}))
        for i, stage in enumerate(stages):
            try:
                desk = next(filter(lambda d: d['_id'] == stage['desk'], desks))
            except StopIteration:
                # if stage has no desk, remove that stage from a list
                logger.warning('Desk not found for stage with id "{}".'.format(stage['_id']))
                stages[i] = None
                continue
            stages[i]['name'] = '{}: {}'.format(desk['name'], stage['name'])
        return tuple(i for i in stages if i)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


from bson import ObjectId
from pytz import timezone
from datetime import timedelta, datetime
from copy import deepcopy
from unittest import mock

import superdesk
import superdesk.signals as signals
from superdesk.errors import SuperdeskApiError
from superdesk.tests import TestCase
from superdesk.utc import utcnow
from superdesk.metadata.item import CONTENT_STATE
from apps.archive.archive import update_image_caption, update_associations
from apps.archive.common import (
    validate_schedule, format_dateline_to_locmmmddsrc, convert_task_attributes_to_objectId,
    get_default_source, set_default_source, get_dateline_city, transtype_metadata
)
from apps.publish.content import publish
from apps.search_providers import register_search_provider, registered_search_providers


NOW = utcnow()


class ArchiveTestCase(TestCase):
    def setUp(self):
        super().setUp()
        search_provider = {'_id': 1, 'source': 'ABC', 'name': 'ABC', 'search_provider': 'ABC'}
        self.app.data.insert('search_providers', [search_provider])
        if not registered_search_providers.get('ABC'):
            register_search_provider('ABC', fetch_endpoint='ABC', label='ABC')

    def test_validate_schedule(self):
        validate_schedule(utcnow() + timedelta(hours=2))

    def test_validate_schedule_at_utc_zero_hours(self):
        validate_schedule((utcnow() + timedelta(days=1)).replace(hour=0, minute=0, second=0, microsecond=0))

    def test_validate_schedule_date_with_datetime_as_string_raises_superdeskApiError(self):
        self.assertRaises(SuperdeskApiError, validate_schedule, "2015-04-27T10:53:48+00:00")

    def test_validate_schedule_date_with_datetime_in_past_raises_superdeskApiError(self):
        self.assertRaises(SuperdeskApiError, validate_schedule, utcnow() + timedelta(hours=-2))

    def _get_located_and_current_utc_ts(self):
        current_ts = utcnow()
        located = {"dateline": "city", "city_code": "Sydney", "state": "NSW", "city": "Sydney", "state_code": "NSW",
                   "country_code": "AU", "tz": "Australia/Sydney", "country": "Australia"}

        current_timestamp = datetime.fromtimestamp(current_ts.timestamp(), tz=timezone(located['tz']))
        if current_timestamp.month == 9:
            formatted_date = 'Sept {}'.format(current_timestamp.strftime('%-d'))
        elif 3 <= current_timestamp.month <= 7:
            formatted_date = current_timestamp.strftime('%B %-d')
        else:
            formatted_date = current_timestamp.strftime('%b %-d')

        return located, formatted_date, current_ts

    def test_format_dateline_to_format_when_only_city_is_present(self):
        located, formatted_date, current_ts = self._get_located_and_current_utc_ts()
        formatted_dateline = format_dateline_to_locmmmddsrc(located, current_ts)
        self.assertEqual(formatted_dateline, 'SYDNEY, %s %s -' % (formatted_date, get_default_source()))

    def test_format_dateline_to_format_when_only_city_and_state_are_present(self):
        located, formatted_date, current_ts = self._get_located_and_current_utc_ts()

        located['dateline'] = "city,state"
        formatted_dateline = format_dateline_to_locmmmddsrc(located, current_ts)
        self.assertEqual(formatted_dateline, 'SYDNEY, NSW, %s %s -' % (formatted_date, get_default_source()))

    def test_format_dateline_to_format_when_only_city_and_country_are_present(self):
        located, formatted_date, current_ts = self._get_located_and_current_utc_ts()

        located['dateline'] = "city,country"
        formatted_dateline = format_dateline_to_locmmmddsrc(located, current_ts)
        self.assertEqual(formatted_dateline, 'SYDNEY, AU, %s %s -' % (formatted_date, get_default_source()))

    def test_format_dateline_to_format_when_city_state_and_country_are_present(self):
        located, formatted_date, current_ts = self._get_located_and_current_utc_ts()

        located['dateline'] = "city,state,country"
        formatted_dateline = format_dateline_to_locmmmddsrc(located, current_ts)
        self.assertEqual(formatted_dateline, 'SYDNEY, NSW, AU, %s %s -' % (formatted_date, get_default_source()))

    def test_if_task_attributes_converted_to_objectid(self):
        doc = {
            'task': {
                'user': '562435231d41c835d7b5fb55',
                'desk': ObjectId("562435241d41c835d7b5fb5d"),
                'stage': 'test',
                'last_authoring_desk': 3245,
                'last_production_desk': None
            }
        }

        convert_task_attributes_to_objectId(doc)
        self.assertIsInstance(doc['task']['user'], ObjectId)
        self.assertEqual(doc['task']['desk'], ObjectId("562435241d41c835d7b5fb5d"))
        self.assertEqual(doc['task']['stage'], 'test')
        self.assertEqual(doc['task']['last_authoring_desk'], 3245)
        self.assertIsNone(doc['task']['last_production_desk'])

    def test_if_metadata_are_transtyped(self):
        """Check that date metadata in extra are transtyped correctly"""
        content_type = {
            "_id": "type_1",
            "label": "test_type",
            "editor": {"test_date_field": {"order": 1, "section": "header"}},
            "schema": {
                "test_date_field": {
                    "type": "date",
                    "required": False,
                    "enabled": True,
                    "nullable": True,
                }
            },
        }
        self.app.data.insert('content_types', [content_type])
        doc = {
            "_id": "transtype_1",
            "profile": "type_1",
            "extra": {"test_date_field": "2019-11-06T00:00:00+0000"},
        }

        transtype_metadata(doc)
        self.assertIsInstance(doc['extra']['test_date_field'], datetime)

    def test_if_no_source_defined_on_desk(self):
        desk = {'name': 'sports'}
        self.app.data.insert('desks', [desk])
        located, formatted_date, current_ts = self._get_located_and_current_utc_ts()
        doc = {
            '_id': '123',
            'task': {
                'desk': desk['_id'],
                'stage': desk['working_stage']
            },
            'dateline': {
                'located': located,
                'date': current_ts
            }
        }

        set_default_source(doc)
        self.assertEqual(doc['source'], get_default_source())
        self.assertEqual(doc['dateline']['source'], get_default_source())
        self.assertEqual(doc['dateline']['text'], 'SYDNEY, %s %s -' % (formatted_date, get_default_source()))

    def test_if_source_defined_on_desk(self):
        source = 'FOO'
        desk = {'name': 'sports', 'source': source}
        self.app.data.insert('desks', [desk])
        located, formatted_date, current_ts = self._get_located_and_current_utc_ts()
        doc = {
            '_id': '123',
            'task': {
                'desk': desk['_id'],
                'stage': desk['working_stage']
            },
            'dateline': {
                'located': located,
                'date': current_ts
            }
        }

        set_default_source(doc)
        self.assertEqual(doc['source'], source)
        self.assertEqual(doc['dateline']['source'], source)
        self.assertEqual(doc['dateline']['text'], 'SYDNEY, %s %s -' % (formatted_date, source))

    def test_if_ingest_provider_source_is_preserved(self):
        desk = {'name': 'sports', 'source': 'FOO'}
        self.app.data.insert('desks', [desk])
        ingest_provider = {'_id': 1, 'source': 'ABC'}
        self.app.data.insert('ingest_providers', [ingest_provider])
        located, formatted_date, current_ts = self._get_located_and_current_utc_ts()
        doc = {
            '_id': '123',
            'task': {
                'desk': desk['_id'],
                'stage': desk['working_stage']
            },
            'dateline': {
                'located': located,
                'date': current_ts
            },
            'ingest_provider': 1
        }

        set_default_source(doc)
        self.assertEqual(doc['source'], 'ABC')
        self.assertEqual(doc['dateline']['source'], 'ABC')
        self.assertEqual(doc['dateline']['text'], 'SYDNEY, %s %s -' % (formatted_date, 'ABC'))

    def test_if_ingest_provider_source_is_not_preserved_for_default_ingest(self):
        desk = {'name': 'sports', 'source': 'FOO'}
        self.app.data.insert('desks', [desk])
        ingest_provider = {'_id': 1, 'source': 'AAP'}
        self.app.data.insert('ingest_providers', [ingest_provider])
        located, formatted_date, current_ts = self._get_located_and_current_utc_ts()
        doc = {
            '_id': '123',
            'task': {
                'desk': desk['_id'],
                'stage': desk['working_stage']
            },
            'dateline': {
                'located': located,
                'date': current_ts
            },
            'ingest_provider': 1
        }

        set_default_source(doc)
        self.assertEqual(doc['source'], 'FOO')
        self.assertEqual(doc['dateline']['source'], 'FOO')
        self.assertEqual(doc['dateline']['text'], 'SYDNEY, %s %s -' % (formatted_date, 'FOO'))

    def test_if_search_provider_source_is_preserved(self):
        desk = {'name': 'sports', 'source': 'FOO'}
        self.app.data.insert('desks', [desk])
        doc = {
            '_id': '123',
            'task': {
                'desk': desk['_id'],
                'stage': desk['working_stage']
            },
            'type': 'picture',
            'ingest_provider': 1
        }

        set_default_source(doc)
        self.assertEqual(doc['source'], 'ABC')

    def test_if_item_has_source_then_search_provider_source_is_not_used(self):
        desk = {'name': 'sports', 'source': 'FOO'}
        self.app.data.insert('desks', [desk])
        doc = {
            '_id': '123',
            'task': {
                'desk': desk['_id'],
                'stage': desk['working_stage']
            },
            'type': 'picture',
            'ingest_provider': 1,
            'source': 'bar'
        }

        set_default_source(doc)
        self.assertEqual(doc['source'], 'bar')

    def test_if_image_caption_is_updated(self):
        body = """
        "body_html" : "<p>test 33</p>\n<!-- EMBED START Image {id: \"embedded9127149191\"} -->\n
        <figure><img src=\"http://localhost:5000/api/upload/58ff025eb611402decdb82e1/raw?_schema=http\" alt=\"aa\" />
        <figcaption>[--description--]</figcaption></figure>\n
        <!-- EMBED END Image {id: \"embedded9127149191\"} -->\n<p>faffaf</p>
        """
        changed_body = """
        "body_html" : "<p>test 33</p>\n<!-- EMBED START Image {id: \"embedded9127149191\"} -->\n
        <figure><img src=\"http://localhost:5000/api/upload/58ff025eb611402decdb82e1/raw?_schema=http\" alt=\"aa\" />
        <figcaption>new caption</figcaption></figure>\n
        <!-- EMBED END Image {id: \"embedded9127149191\"} -->\n<p>faffaf</p>
        """
        body = update_image_caption(body, 'embedded9127149191', 'new caption')
        self.assertEqual(body, changed_body)

    def test_update_associations(self):
        doc = {
            'fields_meta': {
                'body_html': {
                    'draftjsState': [{
                        'entityMap': {
                            '1': {
                                'mutability': 'MUTABLE',
                                'type': 'MEDIA',
                                'data': {
                                    'media': {
                                        'guid': 'guid1',
                                        'type': 'picture',
                                        'alt_text': 'media 1'
                                    }
                                }
                            },
                            '0': {
                                'mutability': 'MUTABLE',
                                'type': 'MEDIA',
                                'data': {
                                    'media': {
                                        'guid': 'guid0',
                                        'type': 'picture',
                                        'alt_text': 'media 0'
                                    }
                                }
                            },
                            '2': {
                                'mutability': 'MUTABLE',
                                'type': 'MEDIA',
                                'data': {
                                    'media': {
                                        'guid': 'guid2',
                                        'type': 'picture',
                                        'alt_text': 'media 2'
                                    }
                                }
                            }
                        }
                    }]}},
            'associations': {
                'editor_1': {'guid': 'guid10', 'type': 'picture10', 'alt_text': 'media 10'},
                'editor_7': {'guid': 'guid7', 'type': 'picture7', 'alt_text': 'media 7'},
                'featuremedia': {'guid': 'guid11', 'type': 'picture11'}}
        }

        update_associations(doc)

        self.assertEqual(len(doc['associations']), 5)
        self.assertEqual(doc['associations']['editor_0'], {'guid': 'guid0', 'type': 'picture', 'alt_text': 'media 0'})
        self.assertEqual(doc['associations']['editor_1'], {'guid': 'guid1', 'type': 'picture', 'alt_text': 'media 1'})
        self.assertEqual(doc['associations']['editor_2'], {'guid': 'guid2', 'type': 'picture', 'alt_text': 'media 2'})
        self.assertEqual(doc['associations']['editor_7'], None)
        self.assertEqual(doc['associations']['featuremedia'], {'guid': 'guid11', 'type': 'picture11'})

    def test_get_dateline_city_None(self):
        self.assertEqual(get_dateline_city(None), '')

    def test_get_dateline_city_located_as_none(self):
        self.assertEqual(get_dateline_city({'located': None}), '')

    def test_get_dateline_city_located_as_none_text_as_none(self):
        self.assertEqual(get_dateline_city({'located': None, 'text': None}), '')

    def test_get_dateline_city_from_text(self):
        self.assertEqual(get_dateline_city({'located': None, 'text': 'Sydney, 9 July AAP'}), 'Sydney')

    def test_get_dateline_city_from_located(self):
        self.assertEqual(get_dateline_city({'located': {'city': 'Melbourne'}, 'text': 'Sydney, 9 July AAP'}),
                         'Melbourne')

    def test_get_dateline_city_from_text_no_city(self):
        self.assertEqual(get_dateline_city({'located': {'city': None}, 'text': 'Sydney, 9 July AAP'}),
                         'Sydney')

    def test_get_dateline_city_from_located_with_country(self):
        self.assertEqual(get_dateline_city({'located': {'country': 'Canada'}, 'text': 'Sydney, 9 July AAP'}),
                         'Sydney')

    def test_get_dateline_city_from_text_with_city_state(self):
        self.assertEqual(get_dateline_city({'located': None, 'text': 'City, State, 9 July AAP'}), 'City, State')

    def test_firstpublished(self):
        """Check that "firstpublihed" field is set correctly

        the test create a story, check firstpublished field, then correct it
        and check again that correction is done and firstpublished has not changed
        """
        archive_service = superdesk.get_resource_service('archive')
        correct_service = superdesk.get_resource_service('archive_correct')
        publish_service = superdesk.get_resource_service('archive_publish')
        item = {'_id': 'foo',
                'guid': 'foo',
                'unique_name': 'foo',
                'type': 'text',
                'state': CONTENT_STATE.SUBMITTED,
                '_current_version': 1,
                'rewrite_of': 'bar'}
        archive_service.create([item])
        with mock.patch.object(publish, 'utcnow', lambda: NOW):
            publish_service.patch('foo', {'body_html': 'original'})
        created = publish_service.find_one(None, _id='foo')
        self.assertEqual(NOW, created['firstpublished'])
        correct_service.patch('foo', {'body_html': 'corrected'})
        # we try to update to check that "firstpublished" is not modified
        # note that utcnow MUST NOT be mocked here, else the test would be pointless
        corrected = publish_service.find_one(None, _id='foo')
        self.assertEqual('corrected', corrected['body_html'])
        self.assertEqual(NOW, corrected['firstpublished'])

    def test_update_signals(self):
        def handler(sender, **kwargs):
            pass

        item_update_mock = mock.create_autospec(handler)
        item_updated_mock = mock.create_autospec(handler)

        signals.item_update.connect(item_update_mock)
        signals.item_updated.connect(item_updated_mock)

        archive_service = superdesk.get_resource_service('archive')
        item = {'_id': 'foo'}
        ids = archive_service.create([item])

        updates = {'foo': 'bar'}
        archive_service.update(ids[0], updates, item)

        updated = item.copy()
        updated.update(updates)
        item_update_mock.assert_called_once_with(archive_service, updates=updates, original=item)
        item_updated_mock.assert_called_once_with(archive_service, item=updated, original=item)

        signals.item_update.disconnect(item_update_mock)
        signals.item_updated.disconnect(item_updated_mock)

    def test_duplicate_signals(self):
        def handler(sender, item, original, operation):
            pass

        duplicate_handler_mock = mock.create_autospec(handler)
        duplicated_handler_mock = mock.create_autospec(handler)

        signals.item_duplicate.connect(duplicate_handler_mock)
        signals.item_duplicated.connect(duplicated_handler_mock)

        archive_service = superdesk.get_resource_service('archive')
        original_item = {
            '_id': 'original',
            'headline': 'original item',
            'language': 'en',
        }
        archive_service.create([original_item])
        original_item = archive_service.find_one(None, _id='original')
        item = deepcopy(original_item)
        item['language'] = 'fr'
        translate_guid = archive_service.duplicate_item(item, operation='translate')

        # `assert_called_once_with` is not used due to a lot of metadata is generated during duplication
        assert duplicate_handler_mock.call_count == 1  # use `assert_called_once` for python>=3.6
        assert duplicate_handler_mock.call_args[0][0] is archive_service
        assert duplicate_handler_mock.call_args[1]['item']['guid'] == translate_guid
        assert duplicate_handler_mock.call_args[1]['original']['_id'] == original_item['_id']
        assert duplicate_handler_mock.call_args[1]['operation'] == 'translate'

        assert duplicated_handler_mock.call_count == 1
        assert duplicated_handler_mock.call_args[0][0] is archive_service
        assert duplicated_handler_mock.call_args[1]['item']['guid'] == translate_guid
        assert duplicated_handler_mock.call_args[1]['original']['_id'] == original_item['_id']
        assert duplicated_handler_mock.call_args[1]['operation'] == 'translate'

        signals.item_duplicate.disconnect(duplicate_handler_mock)
        signals.item_duplicated.disconnect(duplicated_handler_mock)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013-2018 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
import hashlib
import datetime
import re
from superdesk.utc import utc
import json
from copy import deepcopy
from superdesk import etree as sd_etree
from superdesk.io.registry import register_feed_parser
from superdesk.io.feed_parsers import FeedParser
from superdesk.io.iptc import subject_codes
from superdesk.metadata.item import ITEM_TYPE, CONTENT_TYPE, GUID_TAG, Priority
from superdesk.metadata.utils import generate_guid
from superdesk.metadata.item import ITEM_TYPE, CONTENT_TYPE, ITEM_URGENCY, ITEM_PRIORITY
from apps.archive.common import format_dateline_to_locmmmddsrc
from superdesk.utc import get_date
from flask import current_app as app
from superdesk import get_resource_service


logger = logging.getLogger(__name__)


class APMediaFeedParser(FeedParser):

    NAME = 'ap_media'

    label = 'AP Media API'

    direct_copy_properties = ('type', ITEM_URGENCY, 'uri', 'language', 'pubstatus', 'ednote', 'headline',
                              'slugline', 'copyrightnotice')

    # Mapping the received urgensy field to a priority value
    priority_map = {1: Priority.Flash.value,  # Flash
                    2: Priority.Three_Paragraph.value,  # Bulletin
                    3: Priority.Urgent.value,  # Urgent
                    4: Priority.Ordinary.value,  # Routine
                    5: Priority.Ordinary.value,  # Daily
                    6: Priority.Ordinary.value,  # Release at will
                    7: Priority.Ordinary.value,  # Weekday advance
                    8: Priority.Ordinary.value}  # Weekend advance

    slug_code_map = {  # Basketball
                       'BKC-': '15008000', 'BKL-': '15008000', 'BKN-': '15008000',
                       'BKW-': '15008000', 'BKO-': '15008000',
                       # Baseball
                       'BBA-': '15007000', 'BBC-': '15007000', 'BBM-': '15007000', 'BBN-': '15007000',
                       'BBO-': '15007000', 'BBY-': '15007000',
                       # Boxing
                       'BOX-': '15014000',
                       # Motor racing
                       'CAR-': '15039000',
                       # Cycling
                       'CYC-': '15019000',
                       # American Football
                       'FBL-': '15003000', 'FBN-': '15003000', 'FBO-': '15003000',
                       # Figure skating
                       'FIG-': '15025000',
                       # Golf
                       'GLF-': '15027000',
                       # Ice Hockey
                       'HKN-': '15031000', 'HKO-': '15031000',
                       # Horse Racing
                       'RAC-': '15030000',
                       # Soccer
                       'SOC-': '15054000',
                       # Tennis
                       'TEN-': '15065000',
                       # Cricket
                       'CRI-': '15017000',
                       # Rugby league
                       'RGL-': '15048000'}

    RELATED_FIELD = None

    RENDITIONS_MAPPING = {
        'baseImage': 'main',
    }

    def __init__(self):
        super().__init__()

    def can_parse(self, s_json):
        try:
            item = s_json.get('data', {}).get('item') and s_json.get('api_version')

            if item is None:
                return False

            return True
        except Exception:
            pass

        return False

    def datetime(self, string):
        try:
            return datetime.datetime.strptime(string, '%Y-%m-%dT%H:%M:%S+0000')
        except ValueError:
            return datetime.datetime.strptime(string, '%Y-%m-%dT%H:%M:%SZ').replace(tzinfo=utc)

    def categorisation_mapping(self, in_item, item):
        """
        Function that maps categorisation fields may be overloaded if not required
        :param in_item:
        :param item:
        :return:
        """
        if in_item.get('subject'):
            categories = [{'qcode': c.get('code')} for c in in_item.get('subject') if c.get('rels') == ['category']]
            if len(categories):
                item['anpa_category'] = [categories[0]]
                self._map_category_codes(item)

        self._map_sluglines_to_subjects(item=item)

    def _map_category_codes(self, item):
        """Map the category code that has been received to a more palatable value

        :param item:
        :return:
        """
        category_code_map = get_resource_service('vocabularies').find_one(req=None, _id='ap_category_map')
        if category_code_map:
            map = {c['ap_code']: c['category_code'] for c in category_code_map['items'] if c['is_active']}
            for category in item.get('anpa_category', []):
                category['qcode'] = map.get(category.get('qcode').lower(), map.get('default'))

    def _map_sluglines_to_subjects(self, item):
        """The first few characters of the slugline may match AP supplimetal categories this is used to set the subject code.

        :param item:
        :return:
        """
        if len(item.get('slugline', '')) > 4:
            qcode = self.slug_code_map.get(item['slugline'][:4])
            if qcode:
                try:
                    item['subject'] = []
                    item['subject'].append({'qcode': qcode, 'name': subject_codes[qcode]})
                except KeyError:
                    logger.debug("Subject code '%s' not found" % qcode)

    def parse(self, s_json, provider=None):
        in_item = s_json.get('data', {}).get('item')
        nitf_item = s_json.get('nitf', {})
        item = {'guid': in_item.get('altids', {}).get('itemid') + ':' + str(in_item.get('version'))}
        item['source'] = provider.get('source') if provider else 'AP'

        for copy_property in self.direct_copy_properties:
            if in_item.get(copy_property) is not None:
                item[copy_property] = in_item[copy_property]

        if in_item.get('version'):
            item['version'] = in_item['version']

        if in_item.get('versioncreated'):
            item['versioncreated'] = self.datetime(in_item.get('versioncreated'))

        if in_item.get('firstcreated'):
            item['firstcreated'] = self.datetime(in_item.get('firstcreated'))

        if len(in_item.get('infosource', [])):
            item['original_source'] = ','.join([n.get('name') for n in in_item.get('infosource', [])])

        if in_item.get('datelinelocation'):
            cities = app.locators.find_cities()
            # Try to find a single matching city either by city and country or city country and state
            located = [c for c in cities if c['city'] == in_item.get('datelinelocation').get('city') and
                       c['country'] == in_item.get('datelinelocation').get('countryname')]
            if len(located) > 1:
                located = [c for c in cities if c['city'] == in_item.get('datelinelocation').get('city') and
                           c['country'] == in_item.get('datelinelocation').get('countryname') and
                           c['state'] == in_item.get('datelinelocation').get('countryareaname')]
            if len(located) == 1:
                item['dateline'] = dict()
                item['dateline']['located'] = located[0]
                item['dateline']['source'] = provider.get('source')
                item['dateline']['text'] = format_dateline_to_locmmmddsrc(item['dateline']['located'],
                                                                          get_date(item['firstcreated']),
                                                                          provider.get('source'))

        if len(in_item.get('bylines', [])):
            item['byline'] = ','.join([n.get('name') if n.get('name') else n.get('by', '') + (
                ' ({})'.format(n.get('title')) if n.get('title') else '') for n in in_item.get('bylines', [])])
            if item.get('byline').startswith('By '):
                item['byline'] = item['byline'][3:]

        if len(in_item.get('usageterms', [])):
            item['usageterms'] = ', '.join([n for n in in_item.get('usageterms', [])])

        if in_item.get('type') == 'picture':
            if in_item.get('renditions'):
                self._parse_renditions(in_item['renditions'], item, provider)

            if in_item.get('description_caption'):
                item['description_text'] = in_item.get('description_caption')
                item['archive_description'] = in_item.get('description_caption')

            if in_item.get('description_creditline'):
                item['credit'] = in_item.get('description_creditline')

            if in_item.get('photographer', {}).get('name'):
                item['byline'] = in_item.get('photographer', {}).get('name')

        if in_item.get('type') == 'text':
            # Peel off the take key if possible
            if ',' in item['slugline']:
                item['anpa_take_key'] = item['slugline'].split(',')[1]
                item['slugline'] = item['slugline'].split(',')[0]
            if item['slugline'].startswith('BC-'):
                item['slugline'] = item['slugline'][3:]
            if item.get('ednote', '').startswith('Eds:'):
                item['ednote'] = item['ednote'][5:]
            if in_item.get('headline_extended'):
                item['abstract'] = in_item.get('headline_extended')

            self.categorisation_mapping(in_item, item)

            # Map the urgency to urgency and priority
            if in_item.get('urgency'):
                item[ITEM_URGENCY] = int(in_item['urgency']) if in_item['urgency'] <= 5 else 5
                item[ITEM_PRIORITY] = self.priority_map.get(in_item['urgency'], 5)

            if nitf_item.get('body_html'):
                # item['body_html'] = sd_etree.clean_html_str(nitf_item.get('body_html'))
                item['body_html'] = nitf_item.get('body_html').replace('<block id="Main">', '').replace('</block>', '')

        if s_json.get('associations'):
            self._parse_associations(s_json['associations'], item, provider)

        return item

    def _parse_associations(self, associations, item, provider=None):
        related_id = getattr(self, 'RELATED_ID', app.config.get('INGEST_AP_RELATED_ID'))
        if related_id:
            item['associations'] = {}
            for key, raw in associations.items():
                item['associations']['{}--{}'.format(related_id, key)] = self.parse(raw, provider)

    def _parse_renditions(self, renditions, item, provider=None):
        try:
            item['renditions'] = {}
            for dest, src in self.RENDITIONS_MAPPING.items():
                rend = renditions[src]
                item['renditions'][dest] = {
                    'href': rend['href'],
                    'mimetype': rend['mimetype'],
                    'width': rend['width'],
                    'height': rend['height'],
                }
        except KeyError:
            pass


register_feed_parser(APMediaFeedParser.NAME, APMediaFeedParser())



from ..service import ProdApiService


class UsersService(ProdApiService):
    excluded_fields = \
        {
            'user_preferences',
            'session_preferences',
            'password',
        } | ProdApiService.excluded_fields



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import io
import re
import logging
import collections

from flask import json, current_app as app, request
from simplejson.errors import JSONDecodeError
from eve.utils import config

from superdesk.errors import SuperdeskApiError
from superdesk.services import BaseService
from superdesk.notification import push_notification
from apps.dictionaries.resource import DICTIONARY_FILE, DictionaryType
from flask_babel import _


FILE_ID = '_file_id'


logger = logging.getLogger(__name__)


def encode_dict(words_dict):
    return json.dumps(words_dict)


def decode_dict(words_list):
    if isinstance(words_list, dict):
        return words_list
    return json.loads(words_list)


def train(features):
    model = collections.defaultdict(lambda: 1)
    for f in features:
        model[f] += 1
    return model


def words(text):
    return [w.strip() for w in re.findall('[^\n]+', text) if not w.isdigit()]


def add_word(words, word, count):
    """Add word with count to words model.

    :param words: words model
    :param word
    :param count
    """
    words.setdefault(word, 0)
    try:
        words[word] += count
    except TypeError as e:
        logger.error(word)
        logger.error(count)
        raise e


def add_words(nwords, text, val=1):
    for word in words(text):
        add_word(nwords, word, val)


def read(stream):
    return stream.read().decode('utf-8').replace('\ufeff', '')


def merge(doc, words):
    doc.setdefault('content', {})
    for word, count in words.items():
        add_word(doc['content'], word, count)


def fetch_dict(doc):
    """Fetch dictionary content after storing it, from gridfs or json.

    :param doc
    """
    if doc and doc.get(FILE_ID):
        content_file = app.storage.get(doc[FILE_ID])
        content = json.loads(content_file.read())
        return content

    if doc and doc.get('content'):
        return decode_dict(doc['content'])

    return {}


def store_dict(updates, original):
    """Store dictionary content.

    In case it's too big - use gridfs, otherwise just json encode it.

    :param updates
    :param original
    """
    content = updates.pop('content', {})
    if content:
        content_json = json.dumps(content)
        if is_big(content_json):
            content_binary = io.BytesIO(content_json.encode('utf-8'))
            updates[FILE_ID] = app.storage.put(content_binary)
            updates['content'] = None
        else:
            updates['content'] = content_json
            updates[FILE_ID] = None

    if original.get(FILE_ID):
        app.storage.delete(original[FILE_ID])


def is_big(json_data):
    """Test if given json data is too big for mongo document.

    Using 1MB as limit atm.

    :param json_data
    """
    return len(json_data) > 1000000


def read_from_file(doc):
    """Read from plain text file

    One word per line
    UTF-8 encoding
    """
    content = doc.pop(DICTIONARY_FILE)
    if 'text/' not in content.mimetype:
        raise SuperdeskApiError.badRequestError(_('A text dictionary file is required'))
    return train(words(read(content)))


class DictionaryService(BaseService):
    def on_create(self, docs):
        if len(docs) == 1 and 'content' in docs[0]:
            # works around Eve behaviour which creates sub-dict on each "." it finds in keys
            # cf. SDESK-3083
            # We also test request.data because it is not set in behave tests, and they would fail without it
            try:
                docs[0]['content'] = json.loads(request.data.decode('utf-8'))['content']
            except (KeyError, JSONDecodeError, RuntimeError):
                # request.data is not set during tests, so we ignore those errors
                pass

        for doc in docs:
            if self.is_duplicate_dictionary(doc):
                raise SuperdeskApiError.badRequestError(message=_('The dictionary already exists'),
                                                        payload={'name': 'duplicate'})

            self.__set_default(doc)
            self._validate_dictionary(doc)

            if doc.get(DICTIONARY_FILE):
                words = read_from_file(doc)
                merge(doc, words)

            store_dict(doc, {})

    def on_created(self, docs):
        for doc in docs:
            push_notification('dictionary:created', language=doc.get('language_id'))

    def find_one(self, req, **lookup):
        doc = super().find_one(req, **lookup)
        if doc:
            doc['content'] = fetch_dict(doc)
        return doc

    def _validate_dictionary(self, updates, original=None):
        if original is None:
            original = {}
        dict_type = updates.get('type', original.get('type', DictionaryType.DICTIONARY.value))
        if dict_type == DictionaryType.ABBREVIATIONS.value and not updates.get('user', original.get('user')):
            raise SuperdeskApiError.badRequestError(message=_('User is required for the abbreviations dictionary.'),
                                                    payload={'user': 'missing'})

        if original and dict_type != original.get('type', DictionaryType.DICTIONARY.value):
            raise SuperdeskApiError.badRequestError(message=_('The dictionary type cannot be changed.'))

    def get_base_language(self, lang):
        if lang and lang.find('-') > 0:
            return lang.split('-')[0]

    def is_duplicate_dictionary(self, doc):
        return self.find_one(req=None, name=doc['name'],
                             language_id=doc['language_id'],
                             type=doc.get('type', DictionaryType.DICTIONARY.value))

    def get_dictionaries(self, lang):
        """Returns all the active dictionaries.

        If both the language (en-AU)
        and the base language (en) available, it will return the dict with language

        :param lang:
        :return:
        """
        languages = [{'language_id': lang}]
        base_language = self.get_base_language(lang)
        if base_language:
            languages.append({'language_id': base_language})

        lookup = {'$and': [{'$or': languages},
                           {'$or': [{'is_active': {'$exists': 0}}, {'is_active': 'true'}]},
                           {'$or': [{'type': {'$exists': 0}}, {'type': DictionaryType.DICTIONARY.value}]}]}
        dicts = list(self.get(req=None, lookup=lookup))
        langs = [d['language_id'] for d in dicts]

        if base_language and base_language in langs and lang in langs:
            dicts = [d for d in dicts if d['language_id'] != base_language]

        return dicts

    def get_model_for_lang(self, lang):
        """Get model for given language.

        It will use all active dictionaries for given language combined.

        :param lang: language code
        """
        model = {}
        dicts = self.get_dictionaries(lang)

        for _dict in dicts:
            content = fetch_dict(_dict)
            for word, count in content.items():
                add_word(model, word, count)
        return model

    def on_update(self, updates, original):
        if 'content' in updates:
            # works around Eve behaviour which creates sub-dict on each "." it finds in keys
            # cf. SDESK-3083
            try:
                updates['content'] = json.loads(request.data.decode('utf-8'))['content']
            except (KeyError, JSONDecodeError, RuntimeError):
                # request.data is not set during tests, so we ignore those errors
                pass

        user = original.get('user')
        name = updates.get('name')
        language_id = updates.get('language_id')

        # if user is present it means dictonary is personal/abbrevations.
        if user and language_id:
            personal_dictionary = {
                'name': str(user) + ':' + language_id,
                'language_id': language_id,
                'type': original.get('type'),
            }

        if (user and language_id and language_id != original.get('language_id')
                and self.is_duplicate_dictionary(personal_dictionary)):
            raise SuperdeskApiError.badRequestError(message=_('The dictionary already exists'),
                                                    payload={'name': 'duplicate'})
        elif (name and name != original.get('name') and self.find_one(req=None, name=name)):
            raise SuperdeskApiError.badRequestError(message=_('The dictionary already exists'),
                                                    payload={'name': 'duplicate'})

        # parse json list
        if updates.get('content_list'):
            updates['content'] = json.loads(updates.pop('content_list'))

        if 'type' not in original:
            self.__set_default(updates)

        self._validate_dictionary(updates, original)

        # handle manual changes
        if original.get('type', DictionaryType.DICTIONARY.value) == DictionaryType.DICTIONARY.value:
            nwords = fetch_dict(original).copy()
            for word, val in updates.get('content', {}).items():
                if val:
                    add_words(nwords, word, val)
                else:
                    nwords.pop(word, None)

            updates['content'] = nwords

        # handle uploaded file
        if updates.get(DICTIONARY_FILE):
            file_words = read_from_file(updates)
            merge(updates, file_words)

        store_dict(updates, original)

    def on_updated(self, updates, original):
        push_notification('dictionary:updated', language=updates.get('language_id', original.get('language_id')))

    def __set_default(self, doc):
        if 'type' not in doc:
            doc['type'] = DictionaryType.DICTIONARY.value

    def on_fetched_item(self, doc):
        self.__enhance_items([doc])

    def on_fetched(self, docs):
        self.__enhance_items(docs[config.ITEMS])

    def __enhance_items(self, docs):
        for doc in docs:
            self.__set_default(doc)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015, 2016 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import cProfile
import logging
import superdesk
import superdesk.signals as signals
from flask import current_app as app

from superdesk import get_resource_service
from superdesk.celery_task_utils import get_lock_id
from superdesk.lock import lock, unlock
from superdesk.metadata.item import ITEM_STATE, CONTENT_STATE, PUBLISH_SCHEDULE, SCHEDULE_SETTINGS
from apps.archive.common import ITEM_OPERATION, ARCHIVE, insert_into_versions
from apps.legal_archive.commands import import_into_legal_archive
from apps.publish.enqueue.enqueue_corrected import EnqueueCorrectedService
from apps.publish.enqueue.enqueue_killed import EnqueueKilledService
from apps.publish.enqueue.enqueue_published import EnqueuePublishedService
from apps.publish.published_item import PUBLISH_STATE, QUEUE_STATE, PUBLISHED, ERROR_MESSAGE
from bson.objectid import ObjectId
from eve.utils import config, ParsedRequest
from eve.versioning import resolve_document_version
from superdesk.celery_app import celery
from superdesk.utc import utcnow
from superdesk.profiling import ProfileManager
from apps.content import push_content_notification
from superdesk.errors import ConnectionTimeout
from celery.exceptions import SoftTimeLimitExceeded
from superdesk.publish.publish_content import publish


logger = logging.getLogger(__name__)

profile = cProfile.Profile()

UPDATE_SCHEDULE_DEFAULT = {'seconds': 10}

ITEM_PUBLISH = 'publish'
ITEM_CORRECT = 'correct'
ITEM_KILL = 'kill'
ITEM_TAKEDOWN = 'takedown'
ITEM_UNPUBLISH = 'unpublish'

enqueue_services = {
    ITEM_PUBLISH: EnqueuePublishedService(),
    ITEM_CORRECT: EnqueueCorrectedService(),
    ITEM_KILL: EnqueueKilledService(),
    ITEM_TAKEDOWN: EnqueueKilledService(published_state=CONTENT_STATE.RECALLED),
    ITEM_UNPUBLISH: EnqueueKilledService(published_state=CONTENT_STATE.UNPUBLISHED),
}


def get_enqueue_service(operation):
    try:
        enqueue_services[operation].get_filters()
    except KeyError:
        # Hot fix for https://dev.sourcefabric.org/browse/SDESK-3555
        # FIXME: this issue needs investigation and a proper fix.
        logger.error("unexpected operation: {operation}".format(operation=operation))
        operation = "correct"
        enqueue_services[operation].get_filters()
    return enqueue_services[operation]


class EnqueueContent(superdesk.Command):
    """Add published items to ``publish_queue``.

    Example:
    ::

        $ python manage.py publish:enqueue

    """

    def run(self):
        """Fetches items from publish queue as per the configuration, calls the transmit function.
        """
        lock_name = get_lock_id('publish', 'enqueue_published')
        if not lock(lock_name, expire=310):
            logger.info('Enqueue Task: {} is already running.'.format(lock_name))
            return

        try:
            items = self.get_published_items()

            if len(items) > 0:
                self.enqueue_items(items)
        finally:
            unlock(lock_name)

    def enqueue_item(self, published_item):
        """
        Creates the corresponding entries in the publish queue for the given item
        """
        published_item_id = ObjectId(published_item[config.ID_FIELD])
        published_service = get_resource_service(PUBLISHED)
        archive_service = get_resource_service(ARCHIVE)
        published_update = {QUEUE_STATE: PUBLISH_STATE.IN_PROGRESS, 'last_queue_event': utcnow()}
        try:
            logger.info('Queueing item with id: {} and item_id: {}'.format(published_item_id,
                                                                           published_item['item_id']))

            published_item = published_service.find_one(req=None, _id=published_item_id)
            if published_item.get(QUEUE_STATE) != PUBLISH_STATE.PENDING:
                logger.info('Queue State is not pending for published item {}. It is in {}'.
                            format(published_item_id, published_item.get(QUEUE_STATE)))
                return

            if published_item.get(ITEM_STATE) == CONTENT_STATE.SCHEDULED:
                # if scheduled then change the state to published
                # change the `version` and `versioncreated` for the item
                # in archive collection and published collection.
                versioncreated = utcnow()
                item_updates = {'versioncreated': versioncreated, ITEM_STATE: CONTENT_STATE.PUBLISHED}
                resolve_document_version(document=item_updates, resource=ARCHIVE,
                                         method='PATCH',
                                         latest_doc={config.VERSION: published_item.get(config.VERSION, 1)})

                # update the archive collection
                archive_item = archive_service.find_one(req=None, _id=published_item['item_id'])
                archive_service.system_update(published_item['item_id'], item_updates, archive_item)
                # insert into version.
                insert_into_versions(published_item['item_id'], doc=None)
                # update archive history
                app.on_archive_item_updated(item_updates, archive_item, ITEM_PUBLISH)
                # import to legal archive
                import_into_legal_archive.apply_async(countdown=3, kwargs={'item_id': published_item['item_id']})
                logger.info('Modified the version of scheduled item: {}'.format(published_item_id))

                logger.info('Publishing scheduled item_id: {}'.format(published_item_id))
                # update the published collection
                published_update.update(item_updates)
                published_item.update({'versioncreated': versioncreated,
                                       ITEM_STATE: CONTENT_STATE.PUBLISHED,
                                       config.VERSION: item_updates[config.VERSION]})
                # send a notification to the clients
                push_content_notification(
                    [{'_id': str(published_item['item_id']), 'task': published_item.get('task', None)}])
                #  apply internal destinations
                signals.item_published.send(self,
                                            item=archive_service.find_one(req=None, _id=published_item['item_id']))

            published_service.patch(published_item_id, published_update)
            # queue the item for publishing
            try:
                queued = get_enqueue_service(published_item[ITEM_OPERATION]).enqueue_item(published_item, None)
            except KeyError as key_error:
                error_updates = {QUEUE_STATE: PUBLISH_STATE.ERROR, ERROR_MESSAGE: str(key_error)}
                published_service.patch(published_item_id, error_updates)
                logger.exception('No enqueue service found for operation %s', published_item[ITEM_OPERATION])
                raise

            # if the item is queued in the publish_queue then the state is "queued"
            # else the queue state is "queued_not_transmitted"
            queue_state = PUBLISH_STATE.QUEUED if queued else PUBLISH_STATE.QUEUED_NOT_TRANSMITTED
            published_service.patch(published_item_id, {QUEUE_STATE: queue_state})
            logger.info('Queued item with id: {} and item_id: {}'.format(published_item_id, published_item['item_id']))
        except ConnectionTimeout as error:  # recoverable, set state to pending and retry next time
            error_updates = {QUEUE_STATE: PUBLISH_STATE.PENDING, ERROR_MESSAGE: str(error)}
            published_service.patch(published_item_id, error_updates)
            raise
        except SoftTimeLimitExceeded as error:
            error_updates = {QUEUE_STATE: PUBLISH_STATE.PENDING, ERROR_MESSAGE: str(error)}
            published_service.patch(published_item_id, error_updates)
            raise
        except Exception as error:
            error_updates = {QUEUE_STATE: PUBLISH_STATE.ERROR, ERROR_MESSAGE: str(error)}
            published_service.patch(published_item_id, error_updates)
            raise

    def get_published_items(self):
        """
        Get all items with queue state: "pending" that are not scheduled or scheduled time has lapsed.
        """
        query = {
            QUEUE_STATE: PUBLISH_STATE.PENDING,
            '$or': [
                {
                    ITEM_STATE: {'$ne': CONTENT_STATE.SCHEDULED}
                },
                {
                    ITEM_STATE: CONTENT_STATE.SCHEDULED,
                    '{}.utc_{}'.format(SCHEDULE_SETTINGS, PUBLISH_SCHEDULE): {'$lte': utcnow()}
                }
            ]
        }
        request = ParsedRequest()
        request.sort = 'publish_sequence_no'
        request.max_results = 200
        return list(get_resource_service(PUBLISHED).get_from_mongo(req=request, lookup=query))

    def enqueue_items(self, published_items):
        """Creates the corresponding entries in the publish queue for each item

        :param list published_items: the list of items marked for publishing
        """
        failed_items = {}

        for queue_item in published_items:
            try:
                self.enqueue_item(queue_item)
            except Exception:
                logger.exception('Failed to queue item {}'.format(queue_item.get('_id')))
                failed_items[str(queue_item.get('_id'))] = queue_item

        if len(failed_items) > 0:
            logger.error('Failed to publish the following items: {}'.format(failed_items.keys()))


superdesk.command('publish:enqueue', EnqueueContent())


@celery.task(soft_time_limit=300)
def enqueue_published():
    """Pick new items from ``published`` collection and enqueue it."""
    with ProfileManager('publish:enqueue'):
        EnqueueContent().run()



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2019 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from superdesk.resource import Resource
from superdesk.metadata.utils import item_url
from superdesk.auth_server.scopes import Scope


class PlanningResource(Resource):
    url = 'planning'
    item_url = item_url
    item_methods = ['GET']
    resource_methods = ['GET']
    allow_unknown = True
    datasource = {
        'source': 'planning',
        'search_backend': 'elastic',
        'default_sort': [('_updated', -1)]
    }
    privileges = {
        'GET': Scope.PLANNING_READ.name
    }


class EventsResource(Resource):
    url = 'events'
    item_url = item_url
    item_methods = ['GET']
    resource_methods = ['GET']
    allow_unknown = True
    datasource = {
        'source': 'events',
        'search_backend': 'elastic',
        'default_sort': [('_updated', -1)]
    }
    privileges = {
        'GET': Scope.EVENTS_READ.name
    }


class AssignmentsResource(Resource):
    url = 'assignments'
    item_url = item_url
    item_methods = ['GET']
    resource_methods = ['GET']
    allow_unknown = True
    datasource = {
        'source': 'assignments',
        'search_backend': 'elastic',
        'default_sort': [('_updated', -1)]
    }
    privileges = {
        'GET': Scope.ASSIGNMENTS_READ.name
    }


class EventsHistoryResource(Resource):
    url = 'events_history'
    item_url = item_url
    item_methods = ['GET']
    resource_methods = ['GET']
    allow_unknown = True
    datasource = {
        'source': 'events_history',
        'default_sort': [('_updated', -1)]
    }
    privileges = {
        'GET': Scope.EVENTS_READ.name
    }


class EventsFilesResource(Resource):
    url = 'events_files'
    item_url = item_url
    item_methods = ['GET']
    resource_methods = ['GET']
    allow_unknown = True
    datasource = {
        'source': 'events_files',
        'default_sort': [('_updated', -1)]
    }
    privileges = {
        'GET': Scope.EVENTS_READ.name
    }



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013-2018 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import logging
import hashlib
import datetime
import re
import json
from copy import deepcopy

from superdesk.io.registry import register_feed_parser
from superdesk.io.feed_parsers import FeedParser
from superdesk.io.iptc import subject_codes
from superdesk.metadata.item import ITEM_TYPE, CONTENT_TYPE, GUID_TAG
from superdesk.metadata.utils import generate_guid

logger = logging.getLogger(__name__)


class BBCNINJSFeedParser(FeedParser):
    """
    Feed Parser for the BBC's NINJS variant
    """

    # Variant documentation can be found at https://docs.ldrs.org.uk/

    NAME = 'bbc_ninjs'

    label = 'BBC NINJS Variant Parser'

    direct_copy_properties = ('uri', 'language', 'headline', 'urgency', 'pubstatus',
                              'mimetype', 'body_text', 'body_html', 'byline',
                              'description_text')

    # A dictionary of known subjects and their nearest IPTC qcode analogs
    subject_trans = {'policing': '02003000', 'fire-and-rescue-services': '11006001', 'crime': '02001000',
                     'transport': '04015000', 'education': '05000000', 'planning-permission': '11016007',
                     'highways': '04015003', 'waste-management': '04005007', 'public-safety': '11006002',
                     'social-care': '07011000', 'trading-standards': '04008015', 'libraries': '01009000',
                     'rubbish-collection': '04005007', 'recycling': '06009000', 'council-tax': '11013000',
                     'environmental-health': '06005000', 'leisure': '10010000', 'litter': '06005000',
                     'revenue-collection': '04008019', 'health': '07000000', 'dog-fouling': '06000000',
                     'housing': '11016002', 'graffiti': '06000000', 'fly-posting': '06000000',
                     'social-services': '14025004'}

    def __init__(self):
        super().__init__()

    def can_parse(self, s_json):
        try:
            item_ident = re.search('\"total\": *[0-9]+', s_json).group()

            if item_ident is None:
                return False

            results_str = re.search('[0-9]+', item_ident).group()

            return results_str is not None and int(results_str) > 0
        except Exception:
            pass

        return False

    def parse(self, s_json, provider=None):
        parsed = []
        json_items = json.loads(s_json).get('item', [])

        for json_item in json_items:
            parsed.extend(self._parse_item(json_item))

        return parsed

    def _parse_item(self, item):
        items = []

        main = self._parse_main(item)
        items.append(main)

        subjects = main.setdefault('subject', [])
        for subject in item['subject']:
            try:
                parsed_subject = self._parse_subject(subject)
                if parsed_subject is not None:
                    subjects.append(parsed_subject)
            except Exception as ex:
                logger.exception("Exception parsing subject, {}".format(ex))

        associations = main.setdefault('associations', {})
        for association in item['associations']:
            try:
                key, parsed_association = self._parse_association(association)
                associations[key] = parsed_association

                # Also create an image item out of the association rendition
                image = self._create_image(association, main)
                items.append(image)
            except Exception as ex:
                logger.exception("Exception parsing association, {}".format(ex))

        # If there's only a single type we don't need to build a composite type
        if len(items) > 1:
            items.append(self._create_package(items, main))

        return items

    def _create_package(self, items, main):
        """Builds a composite package for the entire article
        :param items:
        :param main: The main article body
        :return: A composite item dict
        """
        package = {
            ITEM_TYPE: CONTENT_TYPE.COMPOSITE,
            'guid': main['guid'] + '-package',
            'versioncreated': main['versioncreated'],
            'firstcreated': main['firstcreated'],
            'headline': main['headline'],
            'groups': [
                {
                    'id': 'root',
                    'role': 'grpRole:NEP',
                    'refs': [{'idRef': 'main'}],
                }, {
                    'id': 'main',
                    'role': 'main',
                    'refs': [],
                }
            ]
        }

        item_references = package['groups'][1]['refs']
        item_references.append({'residRef': main['guid']})

        for item in items:
            if item != main:
                item_references.append({'residRef': item['guid']})

        return package

    def _create_image(self, association, main):
        """Builds an image item from an association
        :param association: The raw association in BBC's ninjs variant
        :param main: The main article body
        :return: A image item dict
        """
        url = association['renditions']['original']['href']
        guid_hash = hashlib.sha1(url.encode('utf8')).hexdigest()

        item = {
            'guid': generate_guid(type=GUID_TAG, id=guid_hash + '-image'),
            ITEM_TYPE: CONTENT_TYPE.PICTURE,
            'versioncreated': main['versioncreated'],
            'firstcreated': main['firstcreated'],
            'headline': association.get('headline', ''),
            'description_text': association.get('description_text', ''),
            'renditions': {
                'baseImage': {
                    'href': url
                }
            }
        }

        return item

    def _parse_association(self, association):
        """Parses a BBC ninjs association
        :param association:
        :return: association dict
        """
        key = association.pop('id')
        # BBC don't use 'featuremedia', they typically use 'featureimage'
        if re.match('^feature', key):
            key = 'featuremedia'

        parsed = deepcopy(association)

        parsed[ITEM_TYPE] = CONTENT_TYPE.PICTURE
        url = association['renditions']['original']['href']
        guid_hash = hashlib.sha1(url.encode('utf8')).hexdigest()
        parsed['guid'] = generate_guid(type=GUID_TAG, id=guid_hash)

        return key, parsed

    def _parse_subject(self, subject):
        """Parses a BBC ninjs subject
        :param subject:
        :return: subject dict
        """
        parsed = {}
        if (subject['lang'] != 'en') or (subject['rel'] != 'category'):
            return

        search_subject = re.sub('^category:', '', subject['code'])

        # Check if we can find the IPTC subject from the custom BBC ones
        if search_subject in self.subject_trans:
            qcode = self.subject_trans[search_subject]
            parsed['qcode'] = qcode
            parsed['name'] = subject_codes[qcode]
            return parsed

        logger.warn("Could not find subject code ({})".format(subject))

    def _parse_main(self, json):
        """Parses the main body of text and metadata
        :param json:
        :return: dict of article metadata and body
        """
        # No GUID is included so generate one from the link
        main = {}

        guid_hash = hashlib.sha1(json['uri'].encode('utf8')).hexdigest()
        main['guid'] = generate_guid(type=GUID_TAG, id=guid_hash)
        # Copy over all attributes which are the same as Superdesk's ninjs variant
        for copy_property in self.direct_copy_properties:
            if json.get(copy_property) is not None:
                main[copy_property] = json[copy_property]

        main['versioncreated'] = self._parse_date(json['versioncreated'])
        main['firstcreated'] = self._parse_date(json['firstcreated'])

        if json.get('embargotime'):
            main['embargo'] = json['embargotime']

        main['type'] = self._convert_type(json['type'])
        return main

    def _parse_date(self, string):
        """Attempts to parse BBC ninjs time in format YYYY-MM-DDTHH:MM:SS
        :param string:
        :return: datetime
        """
        return datetime.datetime.strptime(string, '%Y-%m-%dT%H:%M:%S')

    def _convert_type(self, content_type):
        """Attempts to convert BBC's types to standard ninjs types
        :param content_type:
        :return:
        """
        if content_type == 'image':
            return CONTENT_TYPE.PICTURE
        if content_type == 'story' or content_type == 'advisory':
            return CONTENT_TYPE.TEXT

        logger.error("could not find content type ({}), defaulting to text".format(content_type))

        return CONTENT_TYPE.TEXT


register_feed_parser(BBCNINJSFeedParser.NAME, BBCNINJSFeedParser())



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license


from superdesk import get_resource_service
from superdesk.services import Service
from superdesk.notification import push_notification
from superdesk.errors import SuperdeskApiError
from eve.utils import config
from eve.utils import ParsedRequest
from flask_babel import _
from copy import deepcopy


class ContactsService(Service):

    def get(self, req, lookup):

        # by default the response will have the inactive and not public entries filtered out
        if 'all' not in req.args:
            lookup['is_active'] = True
            lookup['public'] = True
        return super().get(req, lookup)

    def on_create(self, docs):
        for doc in docs:
            self._validate_assignable(doc)

    def on_created(self, docs):
        """
        Send notification to clients that new contact(s) have been created
        :param docs:
        :return:
        """
        push_notification('contacts:create', _id=[doc.get(config.ID_FIELD) for doc in docs])

    def on_update(self, updates, original):
        item = deepcopy(original)
        item.update(updates)
        self._validate_assignable(item)

    def on_updated(self, updates, original):
        """
        Send notifification to clients that a contact has been updated
        :param updates:
        :param original:
        :return:
        """
        push_notification('contacts:update', _id=[original.get(config.ID_FIELD)])

    def on_deleted(self, doc):
        """
        Send a notification to clients that a contact has been deleted
        :param doc:
        :return:
        """
        push_notification('contacts:deleted', _id=[doc.get(config.ID_FIELD)])

    def _validate_assignable(self, contact):
        """Validates a required email address if the contact_type has assignable flag turned on"""

        if not contact or not contact.get('contact_type'):
            return

        types = get_resource_service('vocabularies').find_one(req=None, _id='contact_type')

        if not types:
            return

        contact_type = next((
            item
            for item in (types.get('items') or [])
            if item.get('qcode') == contact.get('contact_type')
        ), None)

        if not contact_type or not contact_type.get('assignable'):
            return

        if not contact.get('contact_email'):
            raise SuperdeskApiError.badRequestError(
                message=_("Contacts of type \"{contact_type}\" must have an email address").format(
                    contact_type=contact_type.get('name')
                )
            )


class OrganisationService(Service):
    def get(self, req, lookup):
        """
        Search for organisation matching the passed q parameter
        :param req:
        :param lookup:
        :return: List of matching organisation strings
        """
        new_req = ParsedRequest()

        q_str = 'organisation:' + '* organisation:'.join(req.args.get('q', '').split()) + '*'
        new_req.args = {'q': q_str, 'default_operator': 'AND', 'projections': '{"organisation": 1}'}
        ret = super().get(new_req, lookup)

        # Remove any duplicate entries from the response
        orgs = []
        de_duped = []
        for d in ret.docs:
            if d.get('organisation') not in orgs:
                orgs.append(d.get('organisation'))
                de_duped.append(d)
        ret.docs = de_duped
        ret.hits['hits']['total'] = len(ret.docs)
        return ret



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015, 2016 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

"""Amazon media storage module."""

from io import BytesIO
from os.path import splitext
from urllib.parse import urlparse
import json
import logging
import time
import unidecode

import boto3
from botocore.client import Config
from eve.io.media import MediaStorage

from superdesk.media.media_operations import download_file_from_url, guess_media_extension
from superdesk.utc import query_datetime
from .mimetype_mixin import MimetypeMixin

logger = logging.getLogger(__name__)
MAX_KEYS = 1000


class AmazonObjectWrapper(BytesIO):

    def __init__(self, s3_object, name, metadata):
        super().__init__()

        s3_body = s3_object['Body']
        blocksize = 65636
        buf = s3_body.read(amt=blocksize)
        while len(buf) > 0:
            self.write(buf)
            buf = s3_body.read(amt=blocksize)

        self.seek(0)
        self.content_type = s3_object['ContentType']
        self.length = int(s3_object['ContentLength'])
        self.name = name
        self.filename = name
        self.metadata = metadata
        self.upload_date = s3_object['LastModified']
        self.md5 = s3_object['ETag'][1:-1]
        self._id = name


class AmazonMediaStorage(MediaStorage, MimetypeMixin):

    def __init__(self, app=None):
        super().__init__(app)
        self.client = boto3.client(
            's3',
            aws_access_key_id=self.app.config['AMAZON_ACCESS_KEY_ID'],
            aws_secret_access_key=self.app.config['AMAZON_SECRET_ACCESS_KEY'],
            region_name=self.app.config.get('AMAZON_REGION'),
            config=Config(signature_version='s3v4'),
            endpoint_url=self.app.config['AMAZON_ENDPOINT_URL'] or None,
        )
        self.user_metadata_header = 'x-amz-meta-'

    def url_for_media(self, media_id, content_type=None):
        return self.app.upload_url(str(media_id))

    def url_for_download(self, media_id, content_type=None):
        return self.app.download_url(str(media_id))

    def _make_s3_safe(self, _id):
        """
        Removes characters from the input _id that may cause issues when using the string as a key in S3 storage.

        See https://docs.aws.amazon.com/AmazonS3/latest/dev/UsingMetadata.html

        :param _id:
        :return:
        """

        def get_translation_table():
            return ''.maketrans({'\\': '',
                                 '{': '',
                                 '^': '',
                                 '}': '',
                                 '%': '',
                                 '`': '',
                                 ']': '',
                                 '>': '',
                                 '[': '',
                                 '~': '',
                                 '<': '',
                                 '#': '',
                                 '|': '',
                                 "'": '',
                                 '"': ''})

        return unidecode.unidecode(str(_id)).translate(get_translation_table())

    def media_id(self, filename, content_type=None, version=True):
        """Get the ``media_id`` path for the given ``filename``.

        if filename doesn't have an extension one is guessed,
        and additional *version* option to have automatic version or not to have,
        or to send a `string` one.
        """
        path = urlparse(filename).path
        file_extension = splitext(path)[1]

        extension = ''
        if not file_extension:
            extension = str(guess_media_extension(content_type)) if content_type else ''

        if version is True:
            # automatic version is set on 15mins granularity.
            mins_granularity = int(int(time.strftime('%M')) / 4) * 4
            version = '%s%s/' % (time.strftime('%Y%m%d%H%m'), mins_granularity)
        elif version is False:
            version = ''
        else:
            version = '%s/' % version.strip('/')

        return '%s%s%s' % (version, self._make_s3_safe(filename), extension)

    def fetch_rendition(self, rendition):
        stream, name, mime = download_file_from_url(rendition.get('href'))
        return stream

    def call(self, method, **kw):
        kw.setdefault('Bucket', self.app.config['AMAZON_CONTAINER_NAME'])
        if 'Key' in kw:
            kw['Key'] = self.get_key(kw['Key'])
        return getattr(self.client, method)(**kw)

    def get_key(self, key):
        subfolder = self.app.config.get('AMAZON_S3_SUBFOLDER', 'false')
        if key and subfolder and subfolder.lower() != 'false':
            key = '%s/%s' % (subfolder.strip('/'), key)
        return key

    def get(self, id_or_filename, resource=None):
        """Open the file given by name or unique id.

        Note that although the returned file is guaranteed to be a File object,
        it might actually be some subclass. Returns None if no file was found.
        """
        id_or_filename = self._make_s3_safe(id_or_filename)
        try:
            obj = self.call('get_object', Key=id_or_filename)
            if obj:
                metadata = self.extract_metadata_from_headers(obj['Metadata'])
                return AmazonObjectWrapper(obj, id_or_filename, metadata)
        except Exception:
            return None
        return None

    def get_all_keys(self):
        """Return the list of all keys from the bucket."""
        all_keys = []
        try:
            for objects in self._get_all_keys_in_batches():
                all_keys.extend(objects)
        except Exception as ex:
            logger.exception(ex)
        return all_keys

    def _get_all_keys_in_batches(self):
        """Return the list of all keys from the bucket in batches."""
        NextMarker = ''
        subfolder = self.app.config.get('AMAZON_S3_SUBFOLDER') or ''
        while True:
            objects = self.call('list_objects', Marker=NextMarker, MaxKeys=MAX_KEYS, Prefix=subfolder)

            if not objects or len(objects.get('Contents', [])) == 0:
                return

            keys = [obj['Key'] for obj in objects.get('Contents', [])]
            NextMarker = keys[-1]
            yield keys

    def extract_metadata_from_headers(self, request_headers):
        headers = {}
        for key, value in request_headers.items():
            if self.user_metadata_header in key:
                new_key = key.split(self.user_metadata_header)[1]
                if(value):
                    try:
                        headers[new_key] = json.loads(value)
                    except Exception as ex:
                        logger.exception(ex)
        return headers

    def put(self, content, filename=None, content_type=None, resource=None, metadata=None, _id=None, version=True,
            folder=None):
        """Save a new file using the storage system, preferably with the name specified.

        If there already exists a file with this name name, the
        storage system may modify the filename as necessary to get a unique
        name. Depending on the storage system, a unique id or the actual name
        of the stored file will be returned. The content type argument is used
        to appropriately identify the file when it is retrieved.

        :param ByteIO content: Data to store in the file object
        :param str filename: Filename used to store the object
        :param str content_type: Content type of the data to be stored
        :param resource: Superdesk resource, i.e. 'upload' or 'download'
        :param metadata: Not currently used with Amazon S3 storage
        :param str _id: ID to be used as the key in the bucket
        :param version: If True the timestamp will be prepended to the key else a string can be used to prepend the key
        :param str folder: The folder to store the object in
        :return str: The ID that was generated for this object
        """
        # XXX: we don't use metadata here as Amazon S3 as a limit of 2048 bytes (keys + values)
        #      and they are anyway stored in MongoDB (and still part of the file). See issue SD-4231
        logger.debug('Going to save file file=%s media=%s ' % (filename, _id))

        # try to determine mimetype on the server
        content_type = self._get_mimetype(content, filename, content_type)

        if not _id:
            _id = self.media_id(filename, content_type=content_type, version=version)

        if folder:
            _id = '%s/%s' % (folder.rstrip('/'), _id)

        found = self._check_exists(_id)
        if found:
            return _id

        kwargs = {}
        acl = self.app.config['AMAZON_OBJECT_ACL']
        if acl:
            # not sure it's really needed here,
            # probably better to turn on/off public-read on the bucket instead
            kwargs['ACL'] = acl

        try:
            self.call(
                'put_object',
                Key=_id,
                Body=content,
                ContentType=content_type,
                **kwargs
            )
            return _id
        except Exception as ex:
            logger.exception(ex)
            raise

    def delete(self, id_or_filename, resource=None):
        id_or_filename = str(id_or_filename)
        del_res = self.call('delete_object', Key=id_or_filename)
        logger.debug('Amazon S3 file deleted %s with status' % id_or_filename, del_res)

    def delete_objects(self, ids):
        """Delete the objects with given list of ids."""
        try:
            delete_parameters = {'Objects': [{'Key': id} for id in ids], 'Quiet': True}
            response = self.call('delete_objects', Delete=delete_parameters)
            if len(response.get('Errors', [])):
                errors = ','.join(['{}:{}'.format(error['Key'], error['Message']) for error in response['Errors']])
                logger.error('Files couldn\'t be deleted: {}'.format(errors))
                return False, errors
            return True, None
        except Exception as ex:
            logger.exception(ex)
            raise

    def exists(self, id_or_filename, resource=None):
        """Test if given name or unique id already exists in storage system."""
        id_or_filename = str(id_or_filename)
        found = self._check_exists(id_or_filename)
        return found

    def _check_exists(self, id_or_filename):
        try:
            self.call('head_object', Key=id_or_filename)
            return True
        except Exception:
            # File not found
            return False

    def remove_unreferenced_files(self, existing_files):
        """Get the files from S3 and compare against existing and delete the orphans."""
        # TODO: Add AMAZON_S3_SUBFOLDER support ref: SDESK-1119
        bucket_files = self.get_all_keys()
        orphan_files = list(set(bucket_files) - existing_files)
        print('There are {} orphan files...'.format(len(orphan_files)))
        for i in range(0, len(orphan_files), MAX_KEYS):
            batch = orphan_files[i:i + MAX_KEYS]
            print('Cleaning %d orphan files...' % len(batch), end='')
            deleted, errors = self.delete_objects(batch)
            if deleted:
                print('done.')
            else:
                print('failed to clean orphans: {}'.format(errors))
        else:
            print('There\'s nothing to clean.')

    def find(self, folder=None, upload_date=None, resource=None):
        """Search for files in the S3 bucket

        Searches for files in the S3 bucket using a combination of folder name and/or upload date
        comparisons. Also uses the `superdesk.utc.query_datetime` method to compare the upload_date provided
        and the upload_date of the file.

        :param str folder: Folder name
        :param dict upload_date: Upload date with comparison operator (i.e. $lt, $lte, $gt or $gte)
        :param resource: The resource type to use
        :return list: List of files that matched the provided parameters
        """

        files = []
        next_marker = ''
        folder = '{}/'.format(folder) if folder else None
        while True:
            result = self.call(
                'list_objects',
                Marker=next_marker,
                MaxKeys=MAX_KEYS,
                Prefix=folder
            )

            if not result or len(result.get('Contents', [])) <= 0:
                break

            objects = result.get('Contents', [])
            for file in objects:
                if upload_date is not None and not query_datetime(file.get('LastModified'), upload_date):
                    continue
                files.append({
                    '_id': file.get('Key'),
                    'filename': file.get('Key'),
                    'upload_date': file.get('LastModified'),
                    'size': file.get('Size'),
                    '_etag': file.get('ETag')
                })

            next_marker = objects[-1]['Key']

        return files

    def getFilename(self, media_id):
        return media_id



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import json
from datetime import date, timedelta
from eve.utils import ParsedRequest
from flask import Flask
from unittest import mock
from unittest.mock import MagicMock
from werkzeug.datastructures import MultiDict
from superdesk import resources
from content_api.tests import ApiTestCase
from superdesk.tests import TestCase
from superdesk.utc import utcnow
from superdesk import get_resource_service


class FakeAuditService():

    def audit_item(self, doc, id):
        return


class FakeAuditResource():
    service = None

    def __init__(self, service):
        self.service = service


class ItemsServiceTestCase(ApiTestCase):
    """Base class for the `items` service tests."""

    def setUp(self):
        self.app = Flask(__name__)
        self.ctx = self.app.test_request_context('/')
        self.ctx.push()
        resources['api_audit'] = FakeAuditResource(FakeAuditService())

    def tearDown(self):
        self.ctx.pop()

    def _get_target_class(self):
        """Return the class under test.

        Make the test fail immediately if the class cannot be imported.
        """
        try:
            from content_api.items import ItemsService
        except ImportError:
            self.fail("Could not import class under test (ItemsService).")
        else:
            return ItemsService

    def _make_one(self, *args, **kwargs):
        """Create a new instance of the class under test."""
        kwargs['datasource'] = 'items'
        return self._get_target_class()(*args, **kwargs)


class CheckForUnknownParamsMethodTestCase(ItemsServiceTestCase):
    """Tests for the _check_for_unknown_params() helper method."""

    def test_does_not_raise_an_error_on_valid_parameters(self):
        request = MagicMock()
        request.args = MultiDict([('sort_by', 'language')])
        instance = self._make_one()

        try:
            instance._check_for_unknown_params(request, ('foo', 'sort_by', 'bar'))
        except Exception as ex:
            self.fail("Exception unexpectedly raised ({})".format(ex))

    def test_raises_correct_error_on_unexpected_parameters(self):
        request = MagicMock()
        request.args = MultiDict([('param_x', 'something')])

        from content_api.errors import UnexpectedParameterError
        instance = self._make_one()

        with self.assertRaises(UnexpectedParameterError) as context:
            instance._check_for_unknown_params(request, ('foo', 'bar'))

        ex = context.exception
        self.assertEqual(ex.payload, 'Unexpected parameter (param_x)')

    def test_raises_descriptive_error_on_filtering_disabled(self):
        request = MagicMock()
        request.args = MultiDict([('q', '{"language": "en"}')])

        from content_api.errors import UnexpectedParameterError
        instance = self._make_one()

        with self.assertRaises(UnexpectedParameterError) as context:
            instance._check_for_unknown_params(
                request, whitelist=(), allow_filtering=False)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            'Unexpected parameter (q)'
        )

    def test_raises_descriptive_error_on_disabled_start_date_filtering(self):
        request = MagicMock()
        request.args = MultiDict([('start_date', '2015-01-01')])

        from content_api.errors import UnexpectedParameterError
        instance = self._make_one()

        with self.assertRaises(UnexpectedParameterError) as context:
            instance._check_for_unknown_params(
                request, whitelist=(), allow_filtering=False)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            'Filtering by date range is not supported when retrieving a '
            'single object (the "start_date" parameter)'
        )

    def test_raises_descriptive_error_on_disabled_end_date_filtering(self):
        request = MagicMock()
        request.args = MultiDict([('end_date', '2015-01-01')])

        from content_api.errors import UnexpectedParameterError
        instance = self._make_one()

        with self.assertRaises(UnexpectedParameterError) as context:
            instance._check_for_unknown_params(
                request, whitelist=(), allow_filtering=False)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            'Filtering by date range is not supported when retrieving a '
            'single object (the "end_date" parameter)'
        )

    def test_raises_correct_error_on_duplicate_parameters(self):
        request = MagicMock()
        request.args = MultiDict([('foo', 'value 1'), ('foo', 'value 2')])

        from content_api.errors import UnexpectedParameterError
        instance = self._make_one()

        with self.assertRaises(UnexpectedParameterError) as context:
            instance._check_for_unknown_params(request, whitelist=('foo',))

        ex = context.exception
        self.assertEqual(
            ex.payload, "Multiple values received for parameter (foo)")


fake_super_get = MagicMock(name='fake super().get')


@mock.patch('content_api.items.service.BaseService.get', fake_super_get)
class GetMethodTestCase(ItemsServiceTestCase):
    """Tests for the get() method."""

    def setUp(self):
        super().setUp()
        fake_super_get.reset_mock()

    @mock.patch('content_api.items.service.ItemsService._check_for_unknown_params')
    def test_correctly_invokes_parameter_validation(self, fake_check_unknown):
        fake_request = MagicMock()
        fake_request.args = MultiDict()
        lookup = {}

        instance = self._make_one()
        instance.get(fake_request, lookup)

        self.assertTrue(fake_check_unknown.called)
        args, kwargs = fake_check_unknown.call_args

        self.assertGreater(len(args), 0)
        self.assertEqual(args[0], fake_request)

        expected_whitelist = sorted([
            'start_date', 'end_date',
            'include_fields', 'exclude_fields',
            'max_results', 'page', 'version', 'where',
            'q', 'default_operator', 'filter',
            'service', 'subject', 'genre', 'urgency',
            'priority', 'type', 'item_source', 'sort'
        ])

        whitelist_arg = kwargs.get('whitelist')
        if whitelist_arg is not None:
            # NOTE: the whitelist argument is converted to a list, because any
            # iterable type is valid, not just lists
            self.assertEqual(sorted(list(whitelist_arg)), expected_whitelist)
        else:
            # whitelist can also be passed as a positional argument
            self.assertGreater(len(args), 1)
            self.assertEqual(sorted(list(args[1])), expected_whitelist)

    @mock.patch('content_api.items.service.ItemsService._set_fields_filter')
    def test_sets_fields_filter_on_request_object(self, fake_set_fields_filter):
        fake_request = MagicMock()
        fake_request.args = MultiDict()
        fake_request.projection = {}
        lookup = {}

        instance = self._make_one()
        instance.get(fake_request, lookup)

        self.assertTrue(fake_set_fields_filter.called)
        args, _ = fake_set_fields_filter.call_args

        self.assertGreater(len(args), 0)
        self.assertEqual(args[0].projection, {})

    def test_provides_request_object_to_superclass_if_not_given(self):
        lookup = {}

        instance = self._make_one()
        instance.get(None, lookup)

        self.assertTrue(fake_super_get.called)
        args, _ = fake_super_get.call_args
        self.assertEqual(len(args), 2)
        self.assertIsInstance(args[0], ParsedRequest)

    def test_raises_correct_error_on_invalid_start_date_parameter(self):
        request = MagicMock()
        request.args = MultiDict([('start_date', '2015-13-35')])
        lookup = {}

        from content_api.errors import BadParameterValueError
        instance = self._make_one()

        with self.assertRaises(BadParameterValueError) as context:
            instance.get(request, lookup)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            ("start_date parameter must be a valid ISO 8601 date (YYYY-MM-DD) "
             "without the time part"))

    def test_raises_correct_error_on_invalid_end_date_parameter(self):
        request = MagicMock()
        request.args = MultiDict([('end_date', '2015-13-35')])
        lookup = {}

        from content_api.errors import BadParameterValueError
        instance = self._make_one()

        with self.assertRaises(BadParameterValueError) as context:
            instance.get(request, lookup)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            ("end_date parameter must be a valid ISO 8601 date (YYYY-MM-DD) "
             "without the time part"))

    def test_raises_correct_error_if_start_date_greater_than_end_date(self):
        request = MagicMock()
        request.args = MultiDict([
            ('start_date', '2015-02-17'),
            ('end_date', '2015-02-16')
        ])
        lookup = {}

        from content_api.errors import BadParameterValueError
        instance = self._make_one()

        with self.assertRaises(BadParameterValueError) as context:
            instance.get(request, lookup)

        ex = context.exception
        self.assertEqual(
            ex.payload, "Start date must not be greater than end date")

    def test_allows_start_and_end_dates_to_be_equal(self):
        request = MagicMock()
        request.args = MultiDict([
            ('start_date', '2010-01-28'),
            ('end_date', '2010-01-28')
        ])
        lookup = {}
        instance = self._make_one()

        try:
            instance.get(request, lookup)
        except Exception as ex:
            self.fail("Exception unexpectedly raised ({})".format(ex))

    def test_includes_given_date_range_into_query_filter_if_given(self):
        request = MagicMock()
        request.args = MultiDict([
            ('start_date', '2012-08-21'),
            ('end_date', '2012-08-26')
        ])
        lookup = {}

        instance = self._make_one()
        instance.get(request, lookup)

        self.assertTrue(fake_super_get.called)
        args, _ = fake_super_get.call_args
        self.assertGreater(len(args), 0)

        filters = json.loads(args[0].args['filter'])['bool']['must']
        date_filter = filters[0].get('range', {}).get('versioncreated', {})
        expected_filter = {
            'gte': '2012-08-21',
            'lte': '2012-08-26'
        }
        self.assertEqual(date_filter, expected_filter)

    @mock.patch('content_api.items.service.utcnow')
    def test_sets_end_date_to_today_if_not_given(self, fake_utcnow):
        request = MagicMock()
        request.args = MultiDict([('start_date', '2012-08-21')])
        lookup = {}

        fake_utcnow.return_value.date.return_value = date(2014, 7, 15)

        instance = self._make_one()
        instance.get(request, lookup)

        self.assertTrue(fake_super_get.called)
        args, _ = fake_super_get.call_args
        self.assertGreater(len(args), 0)

        filters = json.loads(args[0].args['filter'])['bool']['must']
        date_filter = filters[0].get('range', {}).get('versioncreated', {})
        expected_filter = {
            'gte': '2012-08-21',
            'lte': '2014-07-15'
        }
        self.assertEqual(date_filter, expected_filter)

    def test_sets_start_date_equal_to_end_date_if_not_given(self):
        request = MagicMock()
        request.args = MultiDict([('end_date', '2012-08-21')])
        lookup = {}

        instance = self._make_one()
        instance.get(request, lookup)

        self.assertTrue(fake_super_get.called)
        args, _ = fake_super_get.call_args
        self.assertGreater(len(args), 0)

        filters = json.loads(args[0].args['filter'])['bool']['must']
        date_filter = filters[0].get('range', {}).get('versioncreated', {})
        expected_filter = {
            'gte': '2012-08-14',
            'lte': '2012-08-21'
        }
        self.assertEqual(date_filter, expected_filter)

    @mock.patch('content_api.items.service.utcnow')
    def test_sets_end_date_and_start_date_to_today_if_both_not_given(
        self, fake_utcnow
    ):
        request = MagicMock()
        request.args = MultiDict()
        lookup = {}

        fake_utcnow.return_value.date.return_value = date(2014, 7, 15)

        instance = self._make_one()
        instance.get(request, lookup)

        self.assertTrue(fake_super_get.called)
        args, _ = fake_super_get.call_args
        self.assertGreater(len(args), 0)

        filters = json.loads(args[0].args['filter'])['bool']['must']
        date_filter = filters[0].get('range', {}).get('versioncreated', {})
        expected_filter = {
            'gte': '2014-07-08',
            'lte': '2014-07-15'
        }
        self.assertEqual(date_filter, expected_filter)

    def test_creates_correct_query_if_start_and_end_date_are_the_same(self):
        request = MagicMock()
        request.args = MultiDict([
            ('start_date', '2010-09-17'),
            ('end_date', '2010-09-17')]
        )
        lookup = {}

        instance = self._make_one()
        instance.get(request, lookup)

        self.assertTrue(fake_super_get.called)
        args, _ = fake_super_get.call_args
        self.assertGreater(len(args), 0)

        filters = json.loads(args[0].args['filter'])['bool']['must']
        date_filter = filters[0].get('range', {}).get('versioncreated', {})
        expected_filter = {
            'gte': '2010-09-17',
            'lte': '2010-09-17'
        }
        self.assertEqual(date_filter, expected_filter)

    @mock.patch('content_api.items.service.utcnow')
    def test_raises_correct_error_for_start_date_in_future(self, fake_utcnow):
        request = MagicMock()
        request.args = MultiDict([('start_date', '2007-10-31')])
        lookup = {}

        fake_utcnow.return_value.date.return_value = date(2007, 10, 30)

        from content_api.errors import BadParameterValueError
        instance = self._make_one()

        with self.assertRaises(BadParameterValueError) as context:
            instance.get(request, lookup)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            "Start date (2007-10-31) must not be set in the future "
            "(current server date (UTC): 2007-10-30)"
        )

    @mock.patch('content_api.items.service.utcnow')
    def test_raises_correct_error_for_end_date_in_future(self, fake_utcnow):
        request = MagicMock()
        request.args = MultiDict([('end_date', '2007-10-31')])
        lookup = {}

        fake_utcnow.return_value.date.return_value = date(2007, 10, 30)

        from content_api.errors import BadParameterValueError
        instance = self._make_one()

        with self.assertRaises(BadParameterValueError) as context:
            instance.get(request, lookup)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            "End date (2007-10-31) must not be set in the future "
            "(current server date (UTC): 2007-10-30)"
        )

    def test_raises_error_for_invalid_parameter_for_service(self):
        request = MagicMock()
        request.args = MultiDict([('service', '')])
        lookup = {}

        from content_api.errors import BadParameterValueError
        instance = self._make_one()

        with self.assertRaises(BadParameterValueError) as context:
            instance.get(request, lookup)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            'Bad parameter value for Parameter ({})'.format('service')
        )

    def test_set_filter_for_service(self):
        request = MagicMock()
        request.args = MultiDict([('service', 'i')])
        lookup = {}

        instance = self._make_one()
        instance.get(request, lookup)

        self.assertTrue(fake_super_get.called)
        args, _ = fake_super_get.call_args
        self.assertGreater(len(args), 0)

        filters = json.loads(args[0].args['filter'])['bool']['must']
        self.assertEqual(filters[0], {'terms': {'service.code': ['i']}})

    def test_raises_error_for_invalid_parameter_for_urgency(self):
        request = MagicMock()
        request.args = MultiDict([('urgency', '')])
        lookup = {}

        from content_api.errors import BadParameterValueError
        instance = self._make_one()

        with self.assertRaises(BadParameterValueError) as context:
            instance.get(request, lookup)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            'Bad parameter value for Parameter ({})'.format('urgency')
        )

    def test_set_filter_for_urgency(self):
        request = MagicMock()
        request.args = MultiDict([('urgency', 1)])
        lookup = {}

        instance = self._make_one()
        instance.get(request, lookup)

        self.assertTrue(fake_super_get.called)
        args, _ = fake_super_get.call_args
        self.assertGreater(len(args), 0)

        filters = json.loads(args[0].args['filter'])['bool']['must']
        self.assertEqual(filters[0], {'terms': {'urgency': [1]}})

    def test_set_filter_for_q(self):
        expected = '(foo OR bar) AND headline:test'
        request = MagicMock()
        request.args = MultiDict([('q', expected)])
        lookup = {}

        instance = self._make_one()
        instance.get(request, lookup)

        self.assertTrue(fake_super_get.called)
        args, _ = fake_super_get.call_args
        self.assertGreater(len(args), 0)

        q = args[0].args['q']
        self.assertEqual(q, expected)


class ParseIsoDateMethodTestCase(ItemsServiceTestCase):
    """Tests for the _parse_iso_date() helper method."""

    def test_returns_none_if_none_given(self):
        klass = self._get_target_class()
        result = klass._parse_iso_date(None)
        self.assertIsNone(result)

    def test_returns_date_object_on_valid_iso_date_string(self):
        klass = self._get_target_class()
        result = klass._parse_iso_date('2015-05-15')
        self.assertEqual(result, date(2015, 5, 15))

    def test_raises_value_error_on_invalid_iso_date_string(self):
        klass = self._get_target_class()
        with self.assertRaises(ValueError):
            klass._parse_iso_date('5th May 2015')


class SetFieldsFilterMethodTestCase(ItemsServiceTestCase):
    """Tests for the _set_fields_filter() helper method."""

    def test_raises_error_if_requesting_to_exclude_required_field(self):
        request = MagicMock()
        request.args = MultiDict([('exclude_fields', 'uri')])
        request.projection = None

        from content_api.errors import BadParameterValueError
        instance = self._make_one()

        with self.assertRaises(BadParameterValueError) as context:
            instance._set_fields_filter(request)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            'Cannot exclude a content field required by the NINJS format '
            '(uri).'
        )

    def test_raises_error_if_field_whitelist_and_blacklist_both_given(self):
        request = MagicMock()
        request.args = MultiDict([
            ('include_fields', 'language'),
            ('exclude_fields', 'body_text'),
        ])
        request.projection = None

        from content_api.errors import UnexpectedParameterError
        instance = self._make_one()

        with self.assertRaises(UnexpectedParameterError) as context:
            instance._set_fields_filter(request)

        ex = context.exception
        self.assertEqual(
            ex.payload,
            'Cannot both include and exclude content fields at the same time.'
        )

    def test_raises_error_if_whitelisting_unknown_content_field(self):
        request = MagicMock()
        request.args = MultiDict([('include_fields', 'field_x')])
        request.projection = None

        from content_api.errors import BadParameterValueError
        from content_api.items import ItemsResource

        instance = self._make_one()

        fake_schema = {'foo': 'schema_bar'}
        with mock.patch.object(ItemsResource, 'schema', new=fake_schema):  # @UndefinedVariable
            with self.assertRaises(BadParameterValueError) as context:
                instance._set_fields_filter(request)

            ex = context.exception
            self.assertEqual(
                ex.payload, 'Unknown content field to include (field_x).')

    def test_raises_error_if_blacklisting_unknown_content_field(self):
        request = MagicMock()
        request.args = MultiDict([('exclude_fields', 'field_x')])
        request.projection = None

        from content_api.errors import BadParameterValueError
        from content_api.items import ItemsResource

        instance = self._make_one()

        fake_schema = {'foo': 'schema_bar'}
        with mock.patch.object(ItemsResource, 'schema', new=fake_schema):  # @UndefinedVariable
            with self.assertRaises(BadParameterValueError) as context:
                instance._set_fields_filter(request)

            ex = context.exception
            self.assertEqual(
                ex.payload, 'Unknown content field to exclude (field_x).')

    def test_filters_out_blacklisted_fields_if_requested(self):
        request = MagicMock()
        request.args = MultiDict([('exclude_fields', 'language,version')])
        request.projection = None

        instance = self._make_one()
        instance._set_fields_filter(request)

        projection = json.loads(request.projection) if request.projection else {}
        expected_projection = {
            'language': 0,
            'version': 0,
        }
        self.assertEqual(projection, expected_projection)

    def test_filters_out_all_but_whitelisted_fields_if_requested(self):
        request = MagicMock()
        request.args = MultiDict([('include_fields', 'body_text,byline')])
        request.projection = None

        instance = self._make_one()
        instance._set_fields_filter(request)

        projection = json.loads(request.projection) if request.projection else {}
        expected_projection = {
            'body_text': 1,
            'byline': 1,
        }
        self.assertEqual(projection, expected_projection)


fake_super_find_one = MagicMock(name='fake super().find_one')


@mock.patch('content_api.items.service.BaseService.find_one', fake_super_find_one)
class FindOneMethodTestCase(ItemsServiceTestCase):
    """Tests for the find_one() method."""

    def setUp(self):
        super().setUp()
        fake_super_find_one.reset_mock()

    @mock.patch('content_api.items.service.ItemsService._check_for_unknown_params')
    def test_correctly_invokes_parameter_validation(self, fake_check_unknown):
        fake_request = MagicMock()
        fake_request.args = MultiDict()
        lookup = {'_id': 'my_item'}

        instance = self._make_one()
        instance.find_one(fake_request, **lookup)

        self.assertTrue(fake_check_unknown.called)
        args, kwargs = fake_check_unknown.call_args

        self.assertGreater(len(args), 0)
        self.assertEqual(args[0], fake_request)
        self.assertEqual(kwargs.get('allow_filtering'), False)

        expected_whitelist = sorted(['exclude_fields', 'include_fields', 'version'])

        whitelist_arg = kwargs.get('whitelist')
        if whitelist_arg is not None:
            # NOTE: the whitelist argument is converted to a list, because any
            # iterable type is valid, not just lists
            self.assertEqual(sorted(list(whitelist_arg)), expected_whitelist)
        else:
            # whitelist can also be passed as a positional argument
            self.assertGreater(len(args), 1)
            self.assertEqual(sorted(list(args[1])), expected_whitelist)

    @mock.patch('content_api.items.service.ItemsService._set_fields_filter')
    def test_sets_fields_filter_on_request_object(self, fake_set_fields_filter):
        fake_request = MagicMock()
        fake_request.args = MultiDict()
        fake_request.projection = None
        lookup = {}

        instance = self._make_one()
        instance.find_one(fake_request, **lookup)

        self.assertTrue(fake_set_fields_filter.called)
        args, _ = fake_set_fields_filter.call_args

        self.assertGreater(len(args), 0)
        self.assertIs(args[0], fake_request)

    def test_invokes_superclass_method_with_given_arguments(self):
        request = MagicMock()
        request.args = MultiDict()
        lookup = {'_id': 'my_item'}

        instance = self._make_one()
        instance.find_one(request, **lookup)

        self.assertTrue(fake_super_find_one.called)
        args, kwargs = fake_super_find_one.call_args
        self.assertEqual(len(args), 1)
        self.assertIs(args[0], request)
        self.assertEqual(kwargs['_id'], lookup['_id'])

    def test_provides_request_object_to_superclass_if_not_given(self):
        lookup = {'_id': 'my_item'}

        instance = self._make_one()
        instance.find_one(None, **lookup)

        self.assertTrue(fake_super_find_one.called)
        args, _ = fake_super_find_one.call_args
        self.assertEqual(len(args), 1)
        self.assertIsInstance(args[0], ParsedRequest)


class OnFetchedItemMethodTestCase(ItemsServiceTestCase):
    """Tests for the on_fetched_item() method."""

    def setUp(self):
        super().setUp()

        self.app = Flask(__name__)
        self.app.config['CONTENTAPI_URL'] = 'http://content_api.com'
        self.app.config['URLS'] = {'items': 'items_endpoint'}

        self.app_context = self.app.app_context()
        self.app_context.push()

    def tearDown(self):
        self.app_context.pop()
        super().tearDown()

    def test_sets_uri_field_on_fetched_document(self):
        document = {
            '_id': 'item:123',
            'headline': 'a test item'
        }

        instance = self._make_one(datasource='items')
        instance.on_fetched_item(document)

        self.assertEqual(
            document.get('uri'),
            'http://content_api.com/items_endpoint/item%3A123'  # %3A == urlquote(':')
        )

    def test_removes_non_ninjs_content_fields_from_fetched_document(self):
        document = {
            '_id': 'item:123',
            '_etag': '12345abcde',
            '_created': '12345abcde',
            '_updated': '12345abcde',
            'headline': 'breaking news',
            'ancestors': ['item:1234'],
            'subscribers': [],
            '_current_version': '1',
            '_latest_version': '1'
        }

        instance = self._make_one(datasource='items')
        instance.on_fetched_item(document)

        for field in ('_created', '_etag', '_id', '_updated', 'ancestors',
                      'subscribers', '_current_version', '_latest_version'):
            self.assertNotIn(field, document)

    def test_does_not_remove_hateoas_links_from_fetched_document(self):
        document = {
            '_id': 'item:123',
            'headline': 'breaking news',
            '_links': {
                'self': {
                    'href': 'link/to/item/itself',
                    'title': 'Item'
                }
            }
        }

        instance = self._make_one(datasource='items')
        instance.on_fetched_item(document)

        expected_links = {
            'self': {'href': 'link/to/item/itself', 'title': 'Item'}
        }
        self.assertEqual(document.get('_links'), expected_links)


class OnFetchedMethodTestCase(ItemsServiceTestCase):
    """Tests for the on_fetched() method."""

    def setUp(self):
        super().setUp()

        self.app = Flask(__name__)
        self.app.config['CONTENTAPI_URL'] = 'http://content_api.com'
        self.app.config['URLS'] = {'items': 'items_endpoint'}

        self.app_context = self.app.app_context()
        self.app_context.push()
        self.req_context = self.app.test_request_context('items/')
        self.req_context.push()

    def tearDown(self):
        self.req_context.pop()
        self.app_context.pop()
        super().tearDown()

    def test_sets_uri_field_on_all_fetched_documents(self):
        result = {
            '_items': [
                {'_id': 'item:123', 'headline': 'a test item'},
                {'_id': 'item:555', 'headline': 'another item'},
            ]
        }

        instance = self._make_one(datasource='items')
        instance.on_fetched(result)

        documents = result['_items']
        self.assertEqual(
            documents[0].get('uri'),
            'http://content_api.com/items_endpoint/item%3A123'  # %3A == urlquote(':')
        )
        self.assertEqual(
            documents[1].get('uri'),
            'http://content_api.com/items_endpoint/item%3A555'  # %3A == urlquote(':')
        )

    def test_removes_non_ninjs_content_fields_from_all_fetched_documents(self):
        result = {
            '_items': [{
                '_id': 'item:123',
                '_etag': '12345abcde',
                '_created': '12345abcde',
                '_updated': '12345abcde',
                'headline': 'breaking news',
                'ancestors': ['item:1234'],
                'subscribers': [],
                '_current_version': '1',
                '_latest_version': '1'
            }, {
                '_id': 'item:555',
                '_etag': '67890fedcb',
                '_created': '2121abab',
                '_updated': '2121abab',
                'headline': 'good news',
                'ancestors': ['item:5554'],
                'subscribers': [],
                '_current_version': '1',
                '_latest_version': '1'
            }]
        }

        instance = self._make_one(datasource='items')
        instance.on_fetched(result)

        documents = result['_items']
        for doc in documents:
            for field in ('_created', '_etag', '_id', '_updated', 'ancestors',
                          'subscribers', '_current_version', '_latest_version'):
                self.assertNotIn(field, doc)

    def test_does_not_remove_hateoas_links_from_fetched_documents(self):
        result = {
            '_items': [{
                '_id': 'item:123',
                '_etag': '12345abcde',
                '_created': '12345abcde',
                '_updated': '12345abcde',
                'headline': 'breaking news',
                '_links': {
                    'self': {
                        'href': 'link/to/item_123',
                        'title': 'Item'
                    }
                }
            }, {
                '_id': 'item:555',
                '_etag': '67890fedcb',
                '_created': '2121abab',
                '_updated': '2121abab',
                'headline': 'good news',
                '_links': {
                    'self': {
                        'href': 'link/to/item_555',
                        'title': 'Item'
                    }
                }
            }]
        }

        instance = self._make_one(datasource='items')
        instance.on_fetched(result)

        documents = result['_items']

        expected_links = {
            'self': {'href': 'link/to/item_123', 'title': 'Item'}
        }
        self.assertEqual(documents[0].get('_links'), expected_links)

        expected_links = {
            'self': {'href': 'link/to/item_555', 'title': 'Item'}
        }
        self.assertEqual(documents[1].get('_links'), expected_links)

    def test_sets_collection_self_link_to_relative_original_url(self):
        result = {
            '_items': [],
            '_links': {
                'self': {'href': 'foo/bar/baz'}
            }
        }

        request_url = 'items?start_date=1975-12-31#foo'
        with self.app.test_request_context(request_url):
            instance = self._make_one(datasource='items')
            instance.on_fetched(result)

        self_link = result.get('_links', {}).get('self', {}).get('href')
        self.assertEqual(self_link, 'items?start_date=1975-12-31')

    @mock.patch('content_api.items.service.g')
    def test_removes_associated_item_if_subscriber_is_not_entitled(self, fake_g):
        fake_g.get = MagicMock(return_value='test')
        result = {
            '_items': [
                {
                    '_id': 'item:123',
                    'headline': 'a test item',
                    'associations': {
                        'featuremedia': {
                            '_id': 'a1',
                            'subscribers': ['test']
                        }
                    }
                },
                {
                    '_id': 'item:555',
                    'headline': 'another item',
                    'associations': {
                        'featuremedia': {
                            '_id': 'a2',
                            'subscribers': ['test2']
                        }
                    }
                },
            ]
        }

        with self.app.test_request_context():
            instance = self._make_one(datasource='items')
            instance.on_fetched(result)

        self.assertEqual(result['_items'][0]['associations']['featuremedia']['_id'], 'a1')
        self.assertNotIn('subscribers', result['_items'][0]['associations']['featuremedia'])
        self.assertNotIn('featuremedia', result['_items'][1]['associations'])


class GetUriMethodTestCase(ItemsServiceTestCase):
    """Tests for the _get_uri() helper method."""

    def setUp(self):
        super().setUp()

        self.app = Flask(__name__)
        self.app.config['CONTENTAPI_URL'] = 'http://content_api.com'
        self.app.config['URLS'] = {
            'items': 'items_endpoint',
            'packages': 'packages_endpoint'
        }

        self.app_context = self.app.app_context()
        self.app_context.push()

    def tearDown(self):
        self.app_context.pop()
        super().tearDown()

    def test_generates_correct_uri_for_non_composite_items(self):
        document = {
            '_id': 'foo:bar',
            'type': 'picture'
        }

        instance = self._make_one(datasource='items')
        result = instance._get_uri(document)

        self.assertEqual(result, 'http://content_api.com/items_endpoint/foo%3Abar')

    def test_generates_correct_uri_for_composite_items(self):
        document = {
            '_id': 'foo:bar',
            'type': 'composite'
        }

        instance = self._make_one(datasource='items')
        result = instance._get_uri(document)

        self.assertEqual(result, 'http://content_api.com/packages_endpoint/foo%3Abar')


class GetExpiredItemsTestCase(TestCase):
    """Tests for the `get_expired_items` helper method."""

    def setUp(self):
        utc = utcnow()
        expired = utc - timedelta(days=10)
        not_expired = utc - timedelta(days=1)
        self.app.data.insert('items', [
            {'_id': 'a1', '_updated': not_expired, 'type': 'text'},  # Single item, not expired

            {'_id': 'b2', '_updated': expired, 'type': 'text'},  # Single item, expired

            # Evolved from: parent expired, child not expired
            {'_id': 'c3', '_updated': expired, 'type': 'text'},
            {'_id': 'd4', '_updated': not_expired, 'type': 'text', 'evolvedfrom': 'c3', 'ancestors': ['c3']},

            # Evolved from: parent expired, child expired
            {'_id': 'e5', '_updated': expired, 'type': 'text'},
            {'_id': 'f6', '_updated': expired, 'type': 'text', 'evolvedfrom': 'e5', 'ancestors': ['e5']},

            # Multi-branch evolved from,
            {'_id': 'g7', '_updated': expired, 'type': 'text'},
            {'_id': 'h8', '_updated': expired, 'type': 'text', 'evolvedfrom': 'g7', 'ancestors': ['g7']},
            {'_id': 'i9', '_updated': not_expired, 'type': 'text', 'evolvedfrom': 'h8', 'ancestors': ['g7', 'h8']},

            # Multi-branch evolved from
            {'_id': 'j10', '_updated': expired, 'type': 'text'},
            {'_id': 'k11', '_updated': expired, 'type': 'text', 'evolvedfrom': 'j10', 'ancestors': ['j10']},
            {'_id': 'l12', '_updated': expired, 'type': 'text', 'evolvedfrom': 'k11', 'ancestors': ['j10', 'k11']},
        ])
        self.expired_ids = ['b2', 'c3', 'e5', 'f6', 'g7', 'h8', 'j10', 'k11', 'l12']
        self.service = get_resource_service('items')

    def test_get_only_expired_items(self):
        expired_items = []
        for items in self.service.get_expired_items(expiry_days=8):
            expired_items.extend(items)

        self.assertEqual(len(expired_items), 9)

        for item in expired_items:
            self.assertIn(item['_id'], self.expired_ids)

    def test_generator_iteration(self):
        """Tests that the yield generator works for `get_expired_items`

        Ensures that each iteration contains the correct items
        """
        iterations = 0
        for items in self.service.get_expired_items(expiry_days=8, max_results=4):
            iterations += 1
            self.assertLess(iterations, 4)

            num_items = len(items)
            item_ids = [item['_id'] for item in items]

            if iterations == 1:
                self.assertEqual(num_items, 4)
                self.assertEqual(item_ids, ['b2', 'c3', 'e5', 'f6'])
            elif iterations == 2:
                self.assertEqual(num_items, 4)
                self.assertEqual(item_ids, ['g7', 'h8', 'j10', 'k11'])
            elif iterations == 3:
                self.assertEqual(num_items, 1)
                self.assertEqual(item_ids, ['l12'])

        self.assertEqual(iterations, 3)

    def test_get_expired_not_including_children(self):
        expired_items = []
        for items in self.service.get_expired_items(expiry_days=8, include_children=False):
            expired_items.extend(items)

        self.assertEqual(len(expired_items), 5)

        for item in expired_items:
            self.assertIn(item['_id'], ['b2', 'c3', 'e5', 'g7', 'j10'])



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

from __future__ import absolute_import
from PIL import Image
from io import BytesIO
import logging
from copy import deepcopy
from flask import current_app as app
from .media_operations import process_file_from_stream
from .media_operations import crop_image
from .media_operations import download_file_from_url
from .media_operations import process_file
from .media_operations import guess_media_extension
from .image import fix_orientation
from eve.utils import config
from superdesk import get_resource_service
from superdesk.filemeta import set_filemeta
import os


logger = logging.getLogger(__name__)


def generate_renditions(original, media_id, inserted, file_type, content_type,
                        rendition_config, url_for_media, insert_metadata=True,
                        temporary=False):
    """Generate system renditions for given media file id.

    :param BytesIO original: original image byte stream
    :param str media_id: media id
    :param list inserted: list of media ids inserted
    :param str file_type: file_type
    :param str content_type: content type
    :param dict rendition_config: rendition config
    :param url_for_media: function to generate url
    :param bool insert_metadata: boolean to inserted metadata or not. For AWS storage it is false.
    :param bool temporary: put files created in database in "temp" if True
                           files in "temp" folder will be removed after 24 hours
    :return: dict of renditions
    """
    rend = {'href': app.media.url_for_media(media_id, content_type), 'media': media_id, 'mimetype': content_type}
    renditions = {'original': rend}

    if file_type != 'image':
        return renditions

    original = fix_orientation(original)
    img = Image.open(original)
    width, height = img.size
    rend.update({'width': width})
    rend.update({'height': height})

    # remove crop if original is small
    custom_renditions = get_renditions_spec(without_internal_renditions=True)
    # we only want to remove rendition from rendition_config if they are custom ones
    for rendition, crop in custom_renditions.items():
        if rendition not in rendition_config:
            continue
        if not can_generate_custom_crop_from_original(width, height, crop):
            if rendition in config.RENDITIONS['picture']:
                logger.info('image is too small for rendition "{rendition}", but it is an internal one, '
                            'so we keep it'.format(rendition=rendition))
            else:
                del rendition_config[rendition]

    ext = content_type.split('/')[1].lower()
    if ext in ('JPG', 'jpg'):
        ext = 'jpeg'
    elif ext not in ('jpeg', 'gif', 'tiff', 'png'):
        ext = 'png'
    # make baseImage rendition first
    base_image = rendition_config.pop('baseImage', None)
    specs = list(rendition_config.items())
    if base_image:
        specs.insert(0, ('baseImage', base_image))
    base = original
    for rendition, rsize in specs:
        cropping_data = {}
        # reset
        base.seek(0)
        # create the rendition (can be based on ratio or pixels)
        if rsize.get('width') and rsize.get('height') and rendition not in config.RENDITIONS['picture']:  # custom crop
            resized, width, height, cropping_data = _crop_image_center(base, ext, rsize['width'], rsize['height'])
            # we need crop data for original size image
            cropping_data = _get_center_crop(original, rsize['width'], rsize['height'])
        elif rsize.get('width') or rsize.get('height'):
            resized, width, height = _resize_image(base, (rsize.get('width'), rsize.get('height')), ext)
        elif rsize.get('ratio'):
            resized, width, height, cropping_data = _crop_image(base, ext, rsize.get('ratio'))
        rend_content_type = 'image/%s' % ext
        file_name, rend_content_type, metadata = process_file_from_stream(resized, content_type=rend_content_type)
        resized.seek(0)
        folder = 'temp' if temporary else None
        _id = app.media.put(resized, filename=file_name,
                            content_type=rend_content_type,
                            folder=folder,
                            metadata=metadata if insert_metadata else None)
        inserted.append(_id)
        renditions[rendition] = {'href': url_for_media(_id, rend_content_type), 'media': _id,
                                 'mimetype': 'image/%s' % ext, 'width': width, 'height': height}
        # add the cropping data if exist
        renditions[rendition].update(cropping_data)

        # use baseImage for other renditions once we have it
        if rendition == 'baseImage' and width < img.size[0] and height < img.size[1]:
            base = resized
    return renditions


def can_generate_custom_crop_from_original(width, height, crop):
    """Checks whether custom crop can be generated or not

    :param int width: width of original
    :param int height: height of original
    :param dict crop: custom crop data
    :return bool: True if custom crop is within original image dimensions
    """
    if not crop:
        crop = {}

    if crop.get('ratio'):
        return True

    try:
        crop_width = int(crop['width'])
    except (KeyError, TypeError, ValueError):
        crop_width = None

    try:
        crop_height = int(crop['height'])
    except (KeyError, TypeError, ValueError):
        crop_height = None

    if crop_width is None and crop_height is None:
        return False

    if (crop_width is not None and crop_height is None) or (crop_width is None and crop_height is not None):
        return True

    return width >= crop_width and height >= crop_height


def delete_file_on_error(doc, file_id):
    # Don't delete the file if we are on the import from storage flow
    if doc.get('_import', None):
        return
    app.media.delete(file_id)


def _crop_image(content, format, ratio):
    """Crop the image given as a binary stream

    @param content: stream
        The binary stream containing the image
    @param format: str
        The format of the resized image (e.g. png, jpg etc.)
    @param ratio: string, int or float
        Ratio to apply. '16:9', '1:1' etc...
    @return: stream
        Returns the resized image as a binary stream.
    """
    img = Image.open(content)
    width, height = img.size
    if type(ratio) not in [float, int]:
        ratio = ratio.split(':')
        ratio = int(ratio[0]) / int(ratio[1])
    if height * ratio > width:
        new_width = width
        new_height = int(new_width / ratio)
        cropping_data = {
            'CropLeft': 0,
            'CropRight': new_width,
            'CropTop': int((height - new_height) / 2),
            'CropBottom': int((height - new_height) / 2) + new_height,
        }
    else:
        new_width = int(height * ratio)
        new_height = height
        cropping_data = {
            'CropLeft': int((width - new_width) / 2),
            'CropRight': int((width - new_width) / 2) + new_width,
            'CropTop': 0,
            'CropBottom': new_height,
        }
    crop, out = crop_image(content, file_name='crop.for.rendition', cropping_data=cropping_data, image_format=format)
    return out, new_width, new_height, cropping_data


def _get_center_crop(original, width, height):
    image = Image.open(original)
    width = int(width)
    height = int(height)

    width_ratio = image.size[0] / width
    height_ratio = image.size[1] / height

    if width_ratio >= height_ratio:
        dest_width = int(width * height_ratio)
        dest_height = image.size[1]
        offset = int((image.size[0] - dest_width) / 2)
        cropping_data = {
            'CropLeft': offset,
            'CropRight': offset + dest_width,
            'CropTop': 0,
            'CropBottom': dest_height,
        }
    else:
        dest_width = image.size[0]
        dest_height = int(height * width_ratio)
        offset = int((image.size[1] - dest_height) / 2)
        cropping_data = {
            'CropLeft': 0,
            'CropRight': dest_width,
            'CropTop': offset,
            'CropBottom': offset + dest_height,
        }

    return cropping_data


def _crop_image_center(content, format, width, height):
    cropping_data = _get_center_crop(content, width, height)
    crop, out = crop_image(content, file_name='crop.for.rendition', cropping_data=cropping_data, exact_size={
        'width': width,
        'height': height,
    }, image_format=format)
    return out, width, height, cropping_data


def to_int(x):
    """Try to convert x to int."""
    try:
        return int(x)
    except TypeError:
        return x


def _resize_image(content, size, format=None, keepProportions=True):
    """Resize the image given as a binary stream

    @param content: stream
        The binary stream containing the image
    @param format: str
        The format of the resized image (e.g. png, jpg etc.)
    @param size: tuple
        A tuple of width, height
    @param keepProportions: boolean
        If true keep image proportions; it will adjust the resized
        image size.
    @return: stream
        Returns the resized image as a binary stream.
    """
    assert isinstance(size, tuple)
    img = Image.open(content)
    if not format:
        format = img.format
    width, height = img.size
    new_width, new_height = [to_int(x) for x in size]
    if keepProportions:
        if new_width is None and new_height is None:
            raise Exception('size parameter requires at least width or height value')
        # resize with width and height
        if new_width is not None and new_height is not None:
            new_width, new_height = new_width, new_height
            x_ratio = width / new_width
            y_ratio = height / new_height
            if x_ratio > y_ratio:
                new_height = int(height / x_ratio)
            else:
                new_width = int(width / y_ratio)
        # resize with only one dimension
        else:
            original_ratio = width / height
            if new_width is not None:
                new_height = int(int(new_width) / original_ratio)
            else:
                new_width = int(new_height * original_ratio)
    resized = img.resize((new_width, new_height), Image.ANTIALIAS)
    out = BytesIO()
    try:
        resized.save(out, format, quality=85)
    except IOError:
        out = BytesIO()
        resized = resized.convert('RGB')
        resized.save(out, format, quality=85)
    out.seek(0)
    return out, new_width, new_height


def get_renditions_spec(without_internal_renditions=False, no_custom_crops=False):
    """Return the list of the needed renditions.

    It contains the ones defined in settings in `RENDITIONS.picture`
    and the ones defined in vocabularies in `crop_sizes`

    @return: list
        Returns the list of renditions specification.
    """
    rendition_spec = {}
    # renditions required by superdesk
    if not without_internal_renditions:
        rendition_spec = deepcopy(config.RENDITIONS['picture'])

    if not no_custom_crops:
        # load custom renditions sizes
        custom_crops = get_resource_service('vocabularies').find_one(req=None, _id='crop_sizes')
        if custom_crops:
            for crop in custom_crops.get('items'):
                if crop.get('is_active'):
                    rendition_spec[crop['name']] = crop
    return rendition_spec


def update_renditions(item, href, old_item, request_kwargs=None):
    """Update renditions for an item.

    If the old_item has renditions uploaded in to media then the old rendition details are
    assigned to the item, this avoids repeatedly downloading the same image and leaving the media entries orphaned.
    If there is no old_item the original is downloaded and renditions are
    generated.
    :param item: parsed item from source
    :param href: reference to original
    :param old_item: the item that we have already ingested, if it exists
    :return: item with renditions
    """
    inserted = []
    try:
        # If there is an existing set of renditions we keep those
        if old_item:
            media = old_item.get('renditions', {}).get('original', {}).get('media', {})
            if media:
                item['renditions'] = old_item['renditions']
                item['mimetype'] = old_item.get('mimetype')
                item['filemeta'] = old_item.get('filemeta')
                item['filemeta_json'] = old_item.get('filemeta_json')
                return

        content, filename, content_type = download_file_from_url(href, request_kwargs)
        file_type, ext = content_type.split('/')
        metadata = process_file(content, file_type)
        file_guid = app.media.put(content, filename, content_type, metadata)
        inserted.append(file_guid)
        rendition_spec = get_renditions_spec()
        renditions = generate_renditions(content, file_guid, inserted, file_type,
                                         content_type, rendition_spec, app.media.url_for_media)
        item['renditions'] = renditions
        item['mimetype'] = content_type
        set_filemeta(item, metadata)
    except Exception as e:
        logger.exception(e)
        for file_id in inserted:
            app.media.delete(file_id)
        raise


def transfer_renditions(renditions):
    """Transfer the passed renditions to localy held renditions

    Download the renditions as passed and upload them to this instances storage
    Adjust the rendition references
    :param renditions:
    :return: Updated renditions
    """
    if not renditions:
        return
    for rend in iter(renditions.values()):
        if rend.get('media'):
            local = app.media.get(rend['media'])
            if local:
                rend['href'] = app.media.url_for_media(rend['media'], local.content_type)
                continue

        content, filename, content_type = download_file_from_url(rend.get('href'))
        file_type, ext = content_type.split('/')
        metadata = process_file(content, file_type)
        file_guid = app.media.put(content, filename, content_type, metadata)
        rend['href'] = app.media.url_for_media(file_guid, content_type)
        rend['media'] = file_guid


def get_rendition_file_name(rendition):
    """
    Return a file name for the rendition no matter what storage mechanism used by superdesk
    :param rendition:
    :return:
    """
    media_id = str(rendition['media'])
    ext = os.path.splitext(media_id)[-1]
    return media_id.replace('/', '-') + (
        guess_media_extension(rendition.get('mimetype', '')) if not ext else '')



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013 - 2018 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import arrow
from superdesk.io.feed_parsers import FileFeedParser
from superdesk.io.registry import register_feed_parser
from superdesk.errors import ParserError
from superdesk.media.media_operations import process_file_from_stream
from superdesk.media.image import get_meta_iptc
from superdesk.media.iim_codes import TAG
from superdesk.metadata.item import GUID_TAG, ITEM_TYPE, CONTENT_TYPE
from superdesk.metadata import utils
from superdesk.media.renditions import generate_renditions, get_renditions_spec
from superdesk.upload import url_for_media
from superdesk.utc import utcnow
from superdesk import filemeta
from flask import current_app as app
from eve.utils import config
from datetime import datetime
import mimetypes
import logging
import os.path

logger = logging.getLogger(__name__)


class ImageIPTCFeedParser(FileFeedParser):
    """
    Feed Parser which can parse images using IPTC metadata
    """

    NAME = 'image_iptc'
    label = "Image (IPTC metadata)"
    ALLOWED_EXT = mimetypes.guess_all_extensions('image/jpeg')

    DATETIME_FORMAT = '%Y%m%dT%H%M%S%z'

    IPTC_MAPPING = {
        TAG.HEADLINE: 'headline',
        TAG.BY_LINE: 'byline',
        TAG.OBJECT_NAME: 'slugline',
        TAG.CAPTION_ABSTRACT: 'description_text',
        TAG.KEYWORDS: 'keywords',
        TAG.SPECIAL_INSTRUCTIONS: 'ednote',
        TAG.COPYRIGHT_NOTICE: 'copyrightnotice',
        TAG.ORIGINAL_TRANSMISSION_REFERENCE: 'assignment_id',
    }

    def can_parse(self, image_path):
        if not isinstance(image_path, str):
            return False
        return mimetypes.guess_type(image_path)[0] == 'image/jpeg'

    def parse(self, image_path, provider=None):
        try:
            item = self.parse_item(image_path)
            return item
        except Exception as ex:
            raise ParserError.parseFileError(exception=ex, provider=provider)

    def parse_item(self, image_path):
        filename = os.path.basename(image_path)
        content_type = mimetypes.guess_type(image_path)[0]
        guid = utils.generate_guid(type=GUID_TAG)
        item = {'guid': guid,
                'uri': guid,
                config.VERSION: 1,
                ITEM_TYPE: CONTENT_TYPE.PICTURE,
                'mimetype': content_type,
                'versioncreated': utcnow(),
                }
        with open(image_path, 'rb') as f:
            _, content_type, file_metadata = process_file_from_stream(f, content_type=content_type)
            f.seek(0)
            file_id = app.media.put(f, filename=filename, content_type=content_type, metadata=file_metadata)
            filemeta.set_filemeta(item, file_metadata)
            f.seek(0)

            metadata = get_meta_iptc(f)
            f.seek(0)
            self.parse_meta(item, metadata)

            rendition_spec = get_renditions_spec(no_custom_crops=True)
            renditions = generate_renditions(f, file_id, [file_id], 'image',
                                             content_type, rendition_spec, url_for_media)
            item['renditions'] = renditions
        return item

    def parse_date_time(self, date, time):
        if not date or not time:
            return

        datetime_string = '{}T{}'.format(date, time)
        try:
            return datetime.strptime(datetime_string, self.DATETIME_FORMAT)
        except ValueError:
            try:
                arrow.get(datetime_string).datetime
            except ValueError:
                return

    def parse_meta(self, item, metadata):
        datetime_created = self.parse_date_time(metadata.get(TAG.DATE_CREATED), metadata.get(TAG.TIME_CREATED))
        if datetime_created:
            item['firstcreated'] = datetime_created

        # now we map IPTC metadata to superdesk metadata
        for source_key, dest_key in self.IPTC_MAPPING.items():
            try:
                item[dest_key] = metadata[source_key]
            except KeyError:
                continue
        return item


register_feed_parser(ImageIPTCFeedParser.NAME, ImageIPTCFeedParser())



import unittest
from datetime import datetime, timedelta

from apps.rules.routing_rules import Weekdays


class WeekdaysTestCase(unittest.TestCase):

    def test_is_valid_schedule(self):
        self.assertTrue(Weekdays.is_valid_schedule(['MON', 'TUE']))
        self.assertFalse(Weekdays.is_valid_schedule(['MON', 'SRC']))

    def test_is_scheduled(self):
        now = datetime.now()
        day = now.strftime('%A')[:3]

        self.assertFalse(Weekdays.is_scheduled_day(now, []))
        self.assertTrue(Weekdays.is_scheduled_day(now, [day]))
        self.assertFalse(Weekdays.is_scheduled_day(now + timedelta(days=1), [day]))

    def test_weekday_dayname(self):
        now = datetime.now()
        day = now.strftime('%A')[:3].upper()
        self.assertEqual(day, Weekdays.dayname(now))



from flask_babel import _

import superdesk
from apps.archive.common import ARCHIVE
from superdesk import config
from superdesk.errors import SuperdeskApiError
from superdesk.media.video_editor import VideoEditorWrapper
from superdesk.metadata.utils import item_url


TIMELINE_THUMBNAILS_AMOUNT = 60


class VideoEditService(superdesk.Service):
    """
    Use video server for editing video.
    """

    video_editor = VideoEditorWrapper()

    def create(self, docs, **kwargs):
        ids = []
        for doc in docs:
            item = doc.get('item')
            item_id = item[config.ID_FIELD]
            renditions = item['renditions']
            video_id = renditions['original'].get('video_editor_id')
            if not video_id:
                raise SuperdeskApiError.badRequestError(message=_('"video_editor_id" is required'))
            if 'capture' not in doc and 'edit' not in doc:
                raise SuperdeskApiError.badRequestError(message=_('"capture" or "edit" is required'))

            # push task capture preview thumbnail to video server
            if 'capture' in doc:
                capture = doc.pop('capture')
                project = self.video_editor.capture_preview_thumbnail(
                    video_id, position=capture.get('position'),
                    crop=capture.get('crop'),
                    rotate=capture.get('rotate')
                )
                renditions.setdefault('thumbnail', {}).update({
                    'href': project['thumbnails']['preview'].get('url'),
                    'mimetype': project['thumbnails']['preview'].get('mime_type', 'image/png'),
                })
            # push task edit video to video server
            if 'edit' in doc:
                edit = doc.pop('edit')
                project = self.video_editor.edit(video_id, edit)
                renditions.setdefault('original', {}).update({
                    'href': project['url'],
                    'mimetype': project.get('mime_type', 'video/mp4'),
                    'version': project['version'] + 1,
                    'video_editor_id': project.get('_id'),
                })

            original_item = super().find_one(req=None, _id=item_id)
            updates = self.system_update(
                id=item_id,
                updates={'renditions': renditions},
                original=original_item
            )
            item.update(updates)
            ids.append(item_id)
        return ids

    def find_one(self, req, **lookup):
        res = super().find_one(req, **lookup)
        if req is None:
            return res

        video_id = res['renditions']['original']['video_editor_id']
        if req.args.get('action') == 'timeline':
            response = self.video_editor.create_timeline_thumbnails(
                video_id,
                req.args.get('amount', TIMELINE_THUMBNAILS_AMOUNT)
            )
            return {
                config.ID_FIELD: video_id,
                **response
            }
        res['project'] = self.video_editor.find_one(video_id)
        return res

    def on_replace(self, document, original):
        """
        Override to upload thumbnails
        """
        if not document.get('file'):
            return
        # avoid dump file storage
        file = document.pop('file')
        project = original.pop('project')
        data = self.video_editor.upload_preview_thumbnail(project.get('_id'), file)
        document.update(original)
        renditions = document.get('renditions', {})
        renditions.setdefault('thumbnail', {}).update({
            'href': data.get('url'),
            'mimetype': data.get('mimetype'),
        })
        document.update({'renditions': renditions})
        return document


class VideoEditResource(superdesk.Resource):
    item_methods = ['GET', 'PUT']
    resource_methods = ['POST']
    privileges = {
        'POST': ARCHIVE,
        'PUT': ARCHIVE,
    }
    datasource = {
        'projection': {
            'processing': 1,
            'thumbnails': 1,
            'project': 1,
            'renditions': 1,
        }
    }
    item_url = item_url
    schema = {
        'file': {'type': 'file'},
        'item': {
            'type': 'dict',
            'required': False,
            'empty': True,
            'schema': {
                config.ID_FIELD: {
                    'type': 'string',
                    'required': True,
                },
                'renditions': {
                    'type': 'dict',
                    'required': True,
                    'allow_unknown': True,
                    'schema': {
                        'original': {'type': 'dict', 'required': True, 'empty': False}
                    }
                }
            }
        },
        'edit': {
            'type': 'dict',
            'required': False,
            'empty': False,
            'schema': {
                'trim': {
                    'required': False,
                    'regex': '^\\d+\\.?\\d*,\\d+\\.?\\d*$',
                },
                'rotate': {
                    'type': 'integer',
                    'required': False,
                    'allowed': [-270, -180, -90, 90, 180, 270]
                },
                'scale': {
                    'type': 'integer',
                    'required': False
                },
                'crop': {
                    'required': False,
                    'regex': '^\\d+,\\d+,\\d+,\\d+$'
                }
            }
        },
        'capture': {
            'type': 'dict',
            'required': False,
            'empty': False,
            'schema': {
                'position': {
                    'required': False,
                    'type': 'float',
                },
                'rotate': {
                    'type': 'integer',
                    'required': False,
                    'allowed': [-270, -180, -90, 90, 180, 270]
                },
                'scale': {
                    'type': 'integer',
                    'required': False
                },
                'crop': {
                    'required': False,
                    'regex': '^\\d+,\\d+,\\d+,\\d+$'
                }
            }
        },
    }


def init_app(app):
    video_edit_service = VideoEditService(ARCHIVE, backend=superdesk.get_backend())
    VideoEditResource('video_edit', app=app, service=video_edit_service)



# -*- coding: utf-8; -*-
#
# This file is part of Superdesk.
#
# Copyright 2013, 2014, 2015 Sourcefabric z.u. and contributors.
#
# For the full copyright and license information, please see the
# AUTHORS and LICENSE files distributed with this source code, or
# at https://www.sourcefabric.org/superdesk/license

import json

from unittest import mock
from datetime import timedelta

from superdesk.utc import utcnow
from superdesk.tests import TestCase
from superdesk.publish.formatters.ninjs_newsroom_formatter import NewsroomNinjsFormatter
from superdesk.publish import init_app

import planning.assignments as planning_assignments
import planning.planning as planning_planning


@mock.patch('superdesk.publish.subscribers.SubscribersService.generate_sequence_number', lambda self, subscriber: 1)
class NewsroomNinjsFormatterTest(TestCase):
    def setUp(self):
        self.formatter = NewsroomNinjsFormatter()
        init_app(self.app)
        self.maxDiff = None

    def test_products(self):
        self.app.data.insert('content_filters',
                             [{"_id": 3,
                               "content_filter": [{"expression": {"pf": [1], "fc": [2]}}],
                               "name": "soccer-only3"}])
        self.app.data.insert('filter_conditions',
                             [{'_id': 1,
                               'field': 'headline',
                               'operator': 'like',
                               'value': 'test',
                               'name': 'test-1'}])
        self.app.data.insert('filter_conditions',
                             [{'_id': 2,
                               'field': 'urgency',
                               'operator': 'in',
                               'value': '2',
                               'name': 'test-2'}])
        self.app.data.insert('products',
                             [{"_id": 1,
                               "content_filter": {"filter_id": 3, "filter_type": "permitting"},
                               "name": "p-1", "product_type": "api"}])
        self.app.data.insert('vocabularies', [
            {
                "_id": "locators",
                "display_name": "Locators",
                "type": "unmanageable",
                "unique_field": "qcode",
                "items": [
                    {"is_active": True, "name": "NSW", "qcode": "NSW", "state": "New South Wales",
                     "country": "Australia", "world_region": "Oceania", "group": "Australia"},
                ],
            }
        ])
        embargo_ts = (utcnow() + timedelta(days=2))
        article = {
            '_id': 'tag:aap.com.au:20150613:12345',
            'guid': 'tag:aap.com.au:20150613:12345',
            '_current_version': 1,
            'anpa_category': [{'qcode': 'a'}],
            'source': 'AAP',
            'headline': 'This is a test headline',
            'byline': 'joe',
            'slugline': 'slugline',
            'subject': [{'qcode': '02011001', 'name': 'international court or tribunal', 'parent': None},
                        {'qcode': '02011002', 'name': 'extradition'}],
            'anpa_take_key': 'take_key',
            'unique_id': '1',
            'body_html': 'The story body',
            'type': 'text',
            'word_count': '1',
            'priority': 1,
            'profile': 'snap',
            'state': 'published',
            'urgency': 2,
            'pubstatus': 'usable',
            'creditline': 'sample creditline',
            'keywords': ['traffic'],
            'abstract': '<p>sample <b>abstract</b></p>',
            'place': [{'name': 'NSW', 'qcode': 'NSW'}],
            'embargo': embargo_ts,
            'body_footer': '<p>call helpline 999 if you are planning to quit smoking</p>',
            'company_codes': [{'name': 'YANCOAL AUSTRALIA LIMITED', 'qcode': 'YAL', 'security_exchange': 'ASX'}],
            'genre': [{'name': 'Article', 'qcode': 'article'}],
            'flags': {'marked_for_legal': True},
            'extra': {'foo': 'test'},
            'operation': 'publish'
        }
        seq, doc = self.formatter.format(article, {'name': 'Test Subscriber'})[0]
        expected = {
            "guid": "tag:aap.com.au:20150613:12345",
            "version": "1",
            "place": [{"code": "NSW", "name": "New South Wales"}],
            "pubstatus": "usable",
            "body_html": "The story body<p>call helpline 999 if you are planning to quit smoking</p>",
            "type": "text",
            "subject": [{"code": "02011001", "name": "international court or tribunal"},
                        {"code": "02011002", "name": "extradition"}],
            "service": [{"code": "a"}],
            "source": "AAP",
            "headline": "This is a test headline",
            "byline": "joe",
            "urgency": 2,
            "priority": 1,
            "embargoed": embargo_ts.isoformat(),
            "profile": "snap",
            "slugline": "slugline",
            "description_text": "sample abstract",
            "description_html": "<p>sample <b>abstract</b></p>",
            'keywords': ['traffic'],
            'organisation': [{'name': 'YANCOAL AUSTRALIA LIMITED', 'rel': 'Securities Identifier',
                              'symbols': [{'ticker': 'YAL', 'exchange': 'ASX'}]}],
            'genre': [{'name': 'Article', 'code': 'article'}],
            'signal': [{'name': 'Content Warning', 'code': 'cwarn', 'scheme': 'http://cv.iptc.org/newscodes/signal/'}],
            'extra': {'foo': 'test'},
            'charcount': 67,
            'wordcount': 13,
            'readtime': 0,
            'products': [{'code': 1, 'name': 'p-1'}]
        }
        self.assertEqual(json.loads(doc), expected)
        article['urgency'] = 1
        seq, doc = self.formatter.format(article, {'name': 'Test Subscriber'})[0]
        expected = {
            "guid": "tag:aap.com.au:20150613:12345",
            "version": "1",
            "place": [{"code": "NSW", "name": "New South Wales"}],
            "pubstatus": "usable",
            "body_html": "The story body<p>call helpline 999 if you are planning to quit smoking</p>",
            "type": "text",
            "subject": [{"code": "02011001", "name": "international court or tribunal"},
                        {"code": "02011002", "name": "extradition"}],
            "service": [{"code": "a"}],
            "source": "AAP",
            "headline": "This is a test headline",
            "byline": "joe",
            "urgency": 1,
            "priority": 1,
            "embargoed": embargo_ts.isoformat(),
            "profile": "snap",
            "slugline": "slugline",
            "description_text": "sample abstract",
            "description_html": "<p>sample <b>abstract</b></p>",
            'keywords': ['traffic'],
            'organisation': [{'name': 'YANCOAL AUSTRALIA LIMITED', 'rel': 'Securities Identifier',
                              'symbols': [{'ticker': 'YAL', 'exchange': 'ASX'}]}],
            'genre': [{'name': 'Article', 'code': 'article'}],
            'signal': [{'name': 'Content Warning', 'code': 'cwarn', 'scheme': 'http://cv.iptc.org/newscodes/signal/'}],
            'extra': {'foo': 'test'},
            'charcount': 67,
            'wordcount': 13,
            'readtime': 0,
            'products': []
        }
        self.assertEqual(json.loads(doc), expected)

    def test_planning_data(self):
        planning_assignments.init_app(self.app)
        planning_planning.init_app(self.app)

        assignments = [{'coverage_item': 'urn:coverage-id', 'planning_item': 'urn:planning-id'}]
        self.app.data.insert('assignments', assignments)

        article = {
            '_id': 'tag:aap.com.au:20150613:12345',
            'guid': 'tag:aap.com.au:20150613:12345',
            'type': 'text',
            'version': 1,
            'assignment_id': assignments[0]['_id'],
        }

        seq, doc = self.formatter.format(article, {'name': 'Test Subscriber'})[0]
        data = json.loads(doc)

        self.assertEqual('urn:planning-id', data['planning_id'])
        self.assertEqual('urn:coverage-id', data['coverage_id'])

    def test_picture_formatter(self):
        article = {
            "guid": "20150723001158606583",
            "_current_version": 1,
            "slugline": "AMAZING PICTURE",
            "original_source": "AAP",
            "renditions": {
                "viewImage": {
                    "width": 640,
                    "href": "http://localhost:5000/api/upload/55b032041d41c8d278d21b6f/raw?_schema=http",
                    "mimetype": "image/jpeg",
                    "height": 401,
                },
                "original": {
                    "href": "https://one-api.aap.com.au/api/v3/Assets/20150723001158606583/Original/download",
                    "mimetype": "image/jpeg",
                },
            },
            "byline": "MICKEY MOUSE",
            "headline": "AMAZING PICTURE",
            "versioncreated": "2015-07-23T00:15:00.000Z",
            "ednote": "TEST ONLY",
            "type": "picture",
            "pubstatus": "usable",
            "source": "AAP",
            "description": "The most amazing picture you will ever see",
            "guid": "20150723001158606583",
            "body_footer": "<p>call helpline 999 if you are planning to quit smoking</p>",
        }
        seq, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        expected = {
            "byline": "MICKEY MOUSE",
            "renditions": {
                "original": {
                    "href": "https://one-api.aap.com.au/api/v3/Assets/20150723001158606583/Original/download",
                    "mimetype": "image/jpeg",
                },
                "viewImage": {
                    "href": "http://localhost:5000/api/upload/55b032041d41c8d278d21b6f/raw?_schema=http",
                    "mimetype": "image/jpeg",
                    "width": 640,
                    "height": 401,
                }
            },
            "headline": "AMAZING PICTURE",
            "pubstatus": "usable",
            "version": "1",
            "versioncreated": "2015-07-23T00:15:00.000Z",
            "guid": "20150723001158606583",
            "description_html": "The most amazing picture you will ever see<p>call helpline 999 if you are planning to "
            "quit smoking</p>",
            "type": "picture",
            "priority": 5,
            "slugline": "AMAZING PICTURE",
            "ednote": "TEST ONLY",
            "source": "AAP",
            "products": [],
        }
        self.assertEqual(expected, json.loads(doc))
        self.assertIn('viewImage', json.loads(doc).get('renditions'))

    def test_auto_published_item(self):
        article = {
            "guid": "foo",
            "_current_version": 1,
            "slugline": "AMAZING PICTURE",
            "original_source": "AAP",
            "byline": "MICKEY MOUSE",
            "headline": "AMAZING PICTURE",
            "versioncreated": "2015-07-23T00:15:00.000Z",
            "ednote": "TEST ONLY",
            "type": "picture",
            "pubstatus": "usable",
            "source": "AAP",
            "description": "The most amazing picture you will ever see",
            "body_footer": "<p>call helpline 999 if you are planning to quit smoking</p>",
        }
        _, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        processed = json.loads(doc)
        self.assertEqual(processed['guid'], 'foo')
        article['ingest_id'] = 'bar'
        article['ingest_version'] = '7'
        _, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        processed = json.loads(doc)
        self.assertEqual(processed['guid'], 'foo')
        article['auto_publish'] = True
        _, doc = self.formatter.format(article, {"name": "Test Subscriber"})[0]
        processed = json.loads(doc)
        self.assertEqual(processed['guid'], 'bar')
        self.assertEqual(processed['version'], '7')
